"""
get_host_list.py - call the VMDR host list API.
"""

from typing import Union

# from xmltodict import parse
from urllib.parse import urlparse, parse_qs

from ..base.call_api import call_api
from ..auth.token import BasicAuth
from ..exceptions.Exceptions import *
from .data_classes.hosts import VMDRHost, VMDRID
from .data_classes.lists.base_list import BaseList
from ..base import xml_parser


def get_host_list(
    auth: BasicAuth, page_count: Union[int, "all"] = "all", **kwargs
) -> list:
    """
    Get the host list from the VMDR API.
    For a full list of parameters, see the Qualys API documentation: [Qualys API VMPc User Guide](https://cdn2.qualys.com/docs/qualys-api-vmpc-user-guide.pdf)

    Parameters:
        auth (BasicAuth): The authentication object.
        page_count (Union[int, "all"]): The number of pages to get. If "all", get all pages. Defaults to "all".

    :Kwargs:

        action (Optional[str]): The action to perform. Default is 'list'. WARNING: any value you pass is overwritten with 'list'. It is just recognized as valid for the sake of completeness.
        echo_request (Optional[bool]): Whether to show the request. Default is 'False'. ends up being passed to API as 0 or 1.
        show_asset_id (Optional[bool]): Whether to show the asset IDs. Default is 'False'. ends up being passed to API as 0 or 1.
        details (Optional[Union[Literal["Basic", "Basic/AGs", "All", "All/AGs", "None"], None]]): The level of detail to return. Default is 'Basic'. Basic includes host ID, IP, tracking method, DNS, netBIOS, and OS. Basic/AGs includes basic host information plus asset group information. Asset group information includes the asset group ID and title. All shows all basic host information plus last vulnerability and compliance scan dates. All/AGs includes all information plus asset group information: group ID and title. None shows only the host ID (or asset ID if show_asset_id is set to 1 (True)).
        os_pattern (Optional[str]): The OS regex pattern to search for. It follows the PCRE standard, and must be URL-encoded. To match an empty string, use '%5E%24'.
        truncation_limit (Optional[int]): The maximum number of characters to return for each field. Default is 1000, and can be set to 0 for no truncation all the way up to 1,000,000. Past 1M, the API will return an error.
        ips (Optional[str]): Only show assets with certain IP addresses/ranges. Multiple values specified as a comma-separated string. A range is specified with a hyphen. Example: 10.0.0.1-10.0.0.255
        ipv6 (Optional[str]): Only show assets with certain IPv6 addresses/ranges. Multiple values specified as a comma-separated string.
        ag_ids (Optional[str]): Only show assets in certain asset group IDs. Multiple values specified as a comma-separated string. A range is specified with a hyphen. Example: 1-5. A valid ID is required or the API will return an error.
        ag_titles (Optional[str]): Only show assets with specified string in the asset group title. Multiple values specified as a comma-separated string.
        ids (Optional[str]): Only show assets with certain asset IDs. Multiple values specified as a comma-separated string. A range is specified with a hyphen. Example: 1-5. A valid ID is required or the API will return an error.
        id_min (Optional[int]): Only show assets with asset IDs greater than or equal to this value.
        id_max (Optional[int]): Only show assets with asset IDs less than or equal to this value.
        network_ids (Optional[str]): Valid only when the Network Support feature is enabled for the users account). Restrict the request to certain custom network IDs. Multiple network IDs are comma separated.
        compliance_enabled (Optional[bool]): Only show assets with compliance enabled. Default is 'False'. ends up being passed to API as 0 or 1.

        DATE PARAMS

        no_vm_scan_since (Optional[str]): Only show assets that have not been scanned since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        no_compliance_scan_since (Optional[str]): Only show assets that have not had a compliance scan since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_scan_since (Optional[str]): Only show assets that have been scanned since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        compliance_scan_since (Optional[str]): Only show assets that have had a compliance scan since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_processed_before (Optional[str]): Only show assets that have been processed by the VM module before this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_processed_after (Optional[str]): Only show assets that have been processed by the VM module after this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_scan_date_before (Optional[str]): Only show assets that have been scanned by the VM module before this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_scan_date_after (Optional[str]): Only show assets that have been scanned by the VM module after this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_auth_scan_date_before (Optional[str]): Only show assets that have been authenticated scanned by the VM module before this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        vm_auth_scan_date_after (Optional[str]): Only show assets that have been authenticated scanned by the VM module after this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        scap_scan_since (Optional[str]): Only show assets that have been scanned by the SCAP module since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.
        no_scap_scan_since (Optional[str]): Only show assets that have not been scanned by the SCAP module since this date. Formatted as 'YYYY-MM-DD[THH:MM:SSZ]' format UTC/GMT.

        TAG PARAMS

        use_tags (bool): Whether to use tags. Default is 'False'. ends up being passed to API as 0 or 1.
        tag_set_by (Optional[Literal["id","name"]]): search for tags by ID or name. Default is 'id'.
        tag_include_selector (Optional[Literal["any","all"]]): Specified tags will be included if any or all are matched. Default is 'any'. Only valid if use_tags is set to 1 (True).
        tag_exclude_selector (Optional[Literal["any","all"]]): Specified tags will be excluded if any or all are matched. Default is 'any'. Only valid if use_tags is set to 1 (True).
        tag_set_include (Optional[str]): Only show assets with tags that match this string. Multiple values specified as a comma-separated string. Only valid if use_tags is set to 1 (True).
        tag_set_exclude (Optional[str]): Only show assets with tags that do not match this string. Multiple values specified as a comma-separated string. Only valid if use_tags is set to 1 (True).
        show_tags (Optional[bool]): Whether to show tags. Default is 'False'. ends up being passed to API as 0 or 1.

        ARS (Asset Risk Score) PARAMS

        show_ars (Optional[bool]): Whether to show the ARS. Default is 'False'. ends up being passed to API as 0 or 1.
        ars_min (Optional[int]): Only show assets with an ARS greater than or equal to this value.
        ars_max (Optional[int]): Only show assets with an ARS less than or equal to this value.
        show_ars_factors (Optional[bool]): Whether to show the ARS factors. Default is 'False'. ends up being passed to API as 0 or 1.

        TRURISK PARAMS

        show_trurisk (Optional[bool]): Whether to show the TruRisk. Default is 'False'. ends up being passed to API as 0 or 1.
        trurisk_min (Optional[int]): Only show assets with a TruRisk greater than or equal to this value.
        trurisk_max (Optional[int]): Only show assets with a TruRisk less than or equal to this value.
        show_trurisk_factors (Optional[bool]): Whether to show the TruRisk factors. Default is 'False'. ends up being passed to API as 0 or 1.

        CLOUD HOST PARAMS

        host_metadata (Optional[Literal["all", "ec2", "azure", "google"]]): Specify all to list all cloud and non-cloud assets. Specify a service to just get cloud assets for that service plus non-cloud assets.
        host_metadata_fields (Optional[str]): Specify the fields to return for cloud assets. Multiple values specified as a comma-separated string. Only valid if host_metadata is set to a cloud service.
        show_cloud_tags (Optional[bool]): Whether to show cloud tags. Default is 'False'. ends up being passed to API as 0 or 1.
        cloud_tag_fields (Optional[str]): Specify the cloud tag name and value, separated by a colon. Multiple values specified as a comma-separated string. If this is not specified and show_cloud_tags is 1, all tags are returned.

    Returns:

        list[Union[VMDRHost, VMDRID]]: A list of VMDRHost or VMDRID objects.
    """

    responses = BaseList()
    pulled = 0

    # add the action to the kwargs:
    kwargs["action"] = "list"

    if kwargs.get("truncation_limit") and (
        kwargs["truncation_limit"] in [0, "0"]
        and kwargs["details"] not in ["None", None]
    ):
        print(
            "[!] Warning: You have specified to pull all data with no pagination. This is generally not recommended, as it uses lots of resources and take a long time to complete. Please consider specifying a page_count or truncation_limit to avoid this issue."
        )

    while True:
        # make the request:
        response = call_api(
            auth=auth,
            module="vmdr",
            endpoint="get_host_list",
            params=kwargs,
            headers={"X-Requested-With": "qualysdk SDK"},
        )
        if response.status_code != 200:
            print("No data returned.")
            return responses

        xml = xml_parser(response.content)

        if (
            "HOST_LIST" not in xml["HOST_LIST_OUTPUT"]["RESPONSE"]
            and "ID_SET" not in xml["HOST_LIST_OUTPUT"]["RESPONSE"]
        ):
            print("No host list returned.")
            return

        # If details is none, ID_SET will be returned instead of HOST_LIST
        if "ID_SET" in xml["HOST_LIST_OUTPUT"]["RESPONSE"]:
            # check if ID_SET is a list of dicts or a single dict:
            if isinstance(xml["HOST_LIST_OUTPUT"]["RESPONSE"]["ID_SET"]["ID"], dict):
                # if it's a single dict, convert it to a list of dicts:
                xml["HOST_LIST_OUTPUT"]["RESPONSE"]["ID_SET"]["ID"] = [
                    xml["HOST_LIST_OUTPUT"]["RESPONSE"]["ID_SET"]["ID"]
                ]

            for ID in xml["HOST_LIST_OUTPUT"]["RESPONSE"]["ID_SET"]["ID"]:
                # create a VMDRID object and append to responses
                # This code will only run if details=None, so now we just need to check if show_asset_id is set to 1:
                if kwargs.get("show_asset_id"):
                    responses.append(VMDRID(ID=ID, TYPE="asset"))
                else:  # elif not kwargs.get("show_asset_id"):
                    responses.append(VMDRID(ID=ID, TYPE="host"))
        else:
            # HOST_LIST will be returned
            # first, check if xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"]
            # is a list of dicts or a single dict:
            if isinstance(
                xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"], dict
            ):
                # if it's a single dict, convert it to a list of dicts:
                xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"] = [
                    xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"]
                ]

            for host in xml["HOST_LIST_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"]:
                # return a list of VMDRHost objects
                responses.append(VMDRHost.from_dict(host))

        """(
            print(f"Page {pulled+1} of {page_count} complete.")
            if page_count != "all"
            else print(f"Page {pulled+1} complete.")
        )"""
        pulled += 1

        if page_count != "all" and pulled >= page_count:
            print("Page count reached.")
            break

        if "WARNING" in xml["HOST_LIST_OUTPUT"]["RESPONSE"]:
            if "URL" in xml["HOST_LIST_OUTPUT"]["RESPONSE"]["WARNING"]:
                print(
                    f"Pagination detected. Pulling next page from url: {xml['HOST_LIST_OUTPUT']['RESPONSE']['WARNING']['URL']}"
                )
                # get the id_min parameter from the URL to pass into kwargs:
                params = parse_qs(
                    urlparse(
                        xml["HOST_LIST_OUTPUT"]["RESPONSE"]["WARNING"]["URL"]
                    ).query
                )
                kwargs["id_min"] = params["id_min"][0]

            else:
                break
        else:
            break

    return responses



"""
vmdr.py - Contains the functions to upload supported VMDR API pulls to SQL DBs.
"""

from datetime import datetime

from pandas import DataFrame
from sqlalchemy import Connection, types

from .base import upload_data, prepare_dataclass
from ..vmdr.data_classes.lists import BaseList


def upload_vmdr_ags(
    ags: BaseList, cnxn: Connection, override_import_dt: datetime = None
) -> int:
    """
    Upload data from vmdr.get_ag_list() to SQL.

    Args:
        ags (BaseList): The Asset Group List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "TITLE": types.String(),
        "OWNER_ID": types.Integer(),
        "OWNER_USER_ID": types.Integer(),
        "UNIT_ID": types.Integer(),
        "LAST_UPDATE": types.DateTime(),
        "NETWORK_ID": types.Integer(),
        "IP_SET": types.String(),
        "BUSINESS_IMPACT": types.String(),
        "DEFAULT_APPLIANCE_ID": types.Integer(),
        "APPLIANCE_IDS": types.String(),
        "DNS_LIST": types.String(),
        "NETBIOS_LIST": types.String(),
        "HOST_IDS": types.String(),
        "ASSIGNED_USER_IDS": types.String(),
        "ASSIGNED_UNIT_IDS": types.String(),
        "OWNER_USER_NAME": types.String(),
        "CVSS_ENVIRO_CDP": types.String(),
        "CVSS_ENVIRO_TD": types.String(),
        "CVSS_ENVIRO_CR": types.String(),
        "CVSS_ENVIRO_IR": types.String(),
        "CVSS_ENVIRO_AR": types.String(),
        "EC2_IDS": types.String(),
        "COMMENTS": types.String(),
        "DOMAIN_LIST": types.String(),
        "CLOUD_GROUP_NAME": types.String(),
        "CLOUD_INSTANCE_STATE": types.String(),
        "CLOUD_INSTANCE_TYPE": types.String(),
        "CLOUD_IS_SPOT_INSTANCE": types.Boolean(),
        "CLOUD_ARCHITECTURE": types.String(),
        "CLOUD_REGION": types.String(),
        "CLOUD_IMAGE_ID": types.String(),
        "CLOUD_AMI_ID": types.String(),
        "CLOUD_PUBLIC_HOSTNAME": types.String(),
        "CLOUD_PUBLIC_IPV4": types.String(),
        "CLOUD_ACCOUNT_ID": types.String(),
    }

    df = DataFrame([prepare_dataclass(ag) for ag in ags])

    # Upload the data:
    return upload_data(
        df, "vmdr_assetgroups", cnxn, dtype=COLS, override_import_dt=override_import_dt
    )


def upload_vmdr_kb(
    kbs: BaseList, cnxn: Connection, override_import_dt: datetime = None
) -> int:
    """
    Upload data from vmdr.query_kb() to SQL.

    Args:
        kbs (BaseList): The KB List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "QID": types.Integer(),
        "VULN_TYPE": types.String(),
        "SEVERITY_LEVEL": types.Integer(),
        "TITLE": types.String(),
        "CATEGORY": types.String(),
        "LAST_SERVICE_MODIFICATION_DATETIME": types.DateTime(),
        "PUBLISHED_DATETIME": types.DateTime(),
        "CODE_MODIFIED_DATETIME": types.DateTime(),
        "BUGTRAQ_LIST": types.String(),
        "PATCHABLE": types.Boolean(),
        "SOFTWARE_LIST": types.String(),
        "VENDOR_REFERENCE_LIST": types.String(),
        "CVE_LIST": types.String(),
        "DIAGNOSIS": types.String(),
        "SOLUTION": types.String(),
        "CONSEQUENCE": types.String(),
        "CORRELATION": types.String(),
        "CVSS": types.String(),
        "CVSS_V3": types.String(),
        "PCI_REASONS": types.String(),
        "PCI_FLAG": types.Boolean(),
        "THREAT_INTELLIGENCE": types.String(),
        "SUPPORTED_MODULES": types.String(),
        "DISCOVERY": types.String(),
        "IS_DISABLED": types.Boolean(),
        "CHANGE_LOG": types.String(),
        "COMPLIANCE_LIST": types.String(),
        "TECHNOLOGY": types.String(),
        "LAST_CUSTOMIZATION": types.DateTime(),
        "SOLUTION_COMMENT": types.String(),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(kb) for kb in kbs])

    # Upload the data:
    return upload_data(
        df,
        "vmdr_knowledgebase",
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_hosts(
    hosts: BaseList,
    cnxn: Connection,
    is_hld: bool = False,
    override_import_dt: datetime = None,
) -> int:
    """
    Upload data from vmdr.get_host_list() to SQL.

    Args:
        hosts (BaseList): The Host List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        is_hld (bool): If the data is from a Host List Detail pull. You can ignore this.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.

    """

    COLS = {
        "ID": types.Integer(),
        "ASSET_ID": types.Integer(),
        "IP": types.String(),
        "IPV6": types.String(),
        "TRACKING_METHOD": types.String(),
        "DNS": types.String(),
        "NETBIOS": types.String(),
        "OS": types.String(),
        "QG_HOSTID": types.String(),
        "LAST_BOOT": types.DateTime(),
        "LAST_SCAN_DATETIME": types.DateTime(),
        "SERIAL_NUMBER": types.String(),
        "HARDWARE_UUID": types.String(),
        "FIRST_FOUND_DATE": types.DateTime(),
        "LAST_ACTIVITY": types.DateTime(),
        "LAST_ACTIVITY_DATE": types.DateTime(),
        "AGENT_STATUS": types.String(),
        "CLOUD_AGENT_RUNNING_ON": types.String(),
        "TAGS": types.String(),
        "LAST_VULN_SCAN_DATETIME": types.DateTime(),
        "LAST_VULN_SCAN_DATE": types.DateTime(),
        "LAST_VM_SCANNED_DATE": types.DateTime(),
        "LAST_VM_AUTH_SCANNED_DURATION": types.Integer(),
        "LAST_VM_SCANNED_DURATION": types.Integer(),
        "LAST_VM_AUTH_SCANNED_DATE": types.DateTime(),
        "LAST_COMPLIANCE_SCAN_DATETIME": types.DateTime(),
        "LAST_PC_SCANNED_DATE": types.DateTime(),
        "ASSET_GROUP_IDS": types.String(),
        "USER_DEF": types.String(),
        "OWNER": types.String(),
        "CLOUD_PROVIDER": types.String(),
        "CLOUD_SERVICE": types.String(),
        "CLOUD_RESOURCE_ID": types.String(),
        "EC2_INSTANCE_ID": types.String(),
        "CLOUD_GROUP_NAME": types.String(),
        "CLOUD_INSTANCE_STATE": types.String(),
        "CLOUD_INSTANCE_TYPE": types.String(),
        "CLOUD_IS_SPOT_INSTANCE": types.Boolean(),
        "CLOUD_ARCHITECTURE": types.String(),
        "CLOUD_REGION": types.String(),
        "CLOUD_IMAGE_ID": types.String(),
        "CLOUD_AMI_ID": types.String(),
        "CLOUD_PUBLIC_HOSTNAME": types.String(),
        "CLOUD_PUBLIC_IPV4": types.String(),
        "CLOUD_ACCOUNT_ID": types.String(),
        "CLOUD_PROVIDER_TAGS": types.String(),
        "ASSET_RISK_SCORE": types.Integer(),
        "TRURISK_SCORE": types.Integer(),
        "TRURISK_SCORE_FACTORS": types.String(),
        "ASSET_CRITICALITY_SCORE": types.Integer(),
    }

    # Remove METADATA and DNS_DATA from the dataclass. They're parsed out already from dataclass initialization.
    df = DataFrame([prepare_dataclass(host) for host in hosts])

    df.drop(columns=["METADATA", "DNS_DATA", "DETECTION_LIST"], inplace=True)

    # Upload the data, with table depdening on if it is a Host List Detail pull or not:
    return upload_data(
        df,
        "vmdr_hosts_list" if not is_hld else "vmdr_hld_hosts_list",
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_ips(
    ips: BaseList, cnxn: Connection, override_import_dt: datetime = None
) -> int:
    """
    Upload data from vmdr.get_ip_list() to SQL.

    Args:
        ips (BaseList): The IP List to upload from vmdr.get_ip_list().
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "IP": types.String(),
        "TYPE": types.String(),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([str(ip) for ip in ips], columns=["IP"])

    # Add the TYPE column, which shows if it is a single IP or a range:
    df["TYPE"] = df["IP"].apply(lambda x: "Single IP" if "/" not in x else "IP Range")

    # Upload the data:
    return upload_data(
        df, "vmdr_ips", cnxn, dtype=COLS, override_import_dt=override_import_dt
    )


def upload_vmdr_hld(
    hld: BaseList, cnxn: Connection, override_import_dt: datetime = None
) -> int:
    """
    Upload data from vmdr.get_hld() to SQL.

    Args:
        hld (BaseList): The Host List to upload.
        cnxn (Connection): The Connection object to the SQL database.

    Returns:
        int: The number of rows uploaded.

    """

    """
    Get_hld and get_host_list technically return the same data. get_hld just
    includes the DETECTION_LIST attribute. We can use the same upload function
    for the host part, and then snip off the DETECTION_LIST attribute to upload
    to a detections table.
    """

    # Isolate the detection lists. Since the Detection objects themselves
    # have an ID attribute, we can use that to link them back to the host.
    detections = BaseList()
    for host in hld:
        if host.DETECTION_LIST:
            for detection in host.DETECTION_LIST:
                detections.append(detection)

    # upload_vmdr_hosts automatically ignores the DETECTION_LIST attribute,
    # so we can use it here with the is_hld flag set to True to put the hosts
    # in a different table than get_host_list.
    hosts_uploaded = upload_vmdr_hosts(hld, cnxn, is_hld=True)
    print(
        f"Uploaded {hosts_uploaded} hosts to vmdr_hld_hosts_list. Moving to detections..."
    )

    COLS = {
        "UNIQUE_VULN_ID": types.BigInteger(),
        "QID": types.BigInteger(),
        "TYPE": types.String(),
        "SEVERITY": types.Integer(),
        "STATUS": types.String(),
        "SSL": types.Boolean(),
        "RESULTS": types.String(),
        "FIRST_FOUND_DATETIME": types.DateTime(),
        "LAST_FOUND_DATETIME": types.DateTime(),
        "QDS": types.Integer(),
        "QDS_FACTORS": types.String(),
        "TIMES_FOUND": types.Integer(),
        "LAST_TEST_DATETIME": types.DateTime(),
        "LAST_UPDATE_DATETIME": types.DateTime(),
        "IS_IGNORED": types.Boolean(),
        "IS_DISABLED": types.Boolean(),
        "LAST_PROCESSED_DATETIME": types.DateTime(),
        "LAST_FIXED_DATETIME": types.DateTime(),
        "PORT": types.Integer(),
        "PROTOCOL": types.String(),
        "FQDN": types.String(),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(detection) for detection in detections])

    # Set QDS to an integer:
    df["QDS"] = df["QDS"].apply(lambda x: int(x) if x else None)

    # Upload the data:
    return upload_data(
        df,
        "vmdr_hld_detections",
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_scanners(
    scanners: BaseList, cnxn: Connection, override_import_dt: datetime = None
) -> int:
    """
    Upload data from vmdr.get_scanner_list() to SQL.

    Args:
        scanners (BaseList): The Scanner List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "NAME": types.String(),
        "SOFTWARE_VERSION": types.Float(),
        "RUNNING_SCAN_COUNT": types.Integer(),
        "STATUS": types.String(),
        "UUID": types.Uuid(),
        "RUNNING_SLICES_COUNT": types.Integer(),
        "MODEL_NUMBER": types.String(),
        "TYPE": types.String(),
        "SERIAL_NUMBER": types.String(),
        "ACTIVATION_CODE": types.BigInteger(),
        "INTERFACE_SETTINGS": types.String(),
        "PROXY_SETTINGS": types.String(),  # dict.
        "IS_CLOUD_DEPLOYED": types.Boolean(),
        "CLOUD_INFO": types.String(),  # dict. DROP THIS! INFO IS IN BELOW FIELDS:
        "PLATFORM_PROVIDER": types.String(),
        "INSTANCE_ID": types.String(),
        "AMI_ID": types.String(),
        "ACCOUNT_ID": types.String(),
        "INSTANCE_REGION": types.String(),
        "INSTANCE_AVAIBILITY_ZONE": types.String(),
        "INSTANCE_ZONE_TYPE": types.String(),
        "INSTANCE_VPC_ID": types.String(),
        "INSTANCE_SUBNET_ID": types.String(),
        "INSTANCE_ADDRESS_PRIVATE": types.String(),
        "INSTANCE_ADDRESS_PUBLIC": types.String(),
        "HOSTNAME_PRIVATE": types.String(),
        "SECURITY_GROUPS": types.String(),
        "API_PROXY_SETTINGS": types.String(),
        "VLANS": types.String(),  # dict.
        "STATIC_ROUTES": types.String(),  # dict.
        "ML_LATEST": types.String(),
        "ML_VERSION": types.String(),
        "VULNSIGS_LATEST": types.String(),
        "VULNSIGS_VERSION": types.String(),
        "ASSET_GROUP_COUNT": types.Integer(),
        "ASSET_GROUP_LIST": types.String(),  # BaseList
        "LAST_UPDATED_DATE": types.DateTime(),
        "POLLING_INTERVAL": types.Integer(),
        "USER_LOGIN": types.String(),
        "HEARTBEATS_MISSED": types.Integer(),
        "SS_CONNECTIION": types.String(),
        "SS_LAST_CONNECTED": types.DateTime(),
        "USER_LIST": types.String(),  # dict
        "UPDATED": types.String(),
        "COMMENTS": types.String(),
        "MAX_CAPACITY_UNITS": types.Float(),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(scanner) for scanner in scanners])

    # Drop the CLOUD_INFO and EC2_INFO columns:
    df.drop(columns=["CLOUD_INFO"], inplace=True)

    # Upload the data:
    return upload_data(
        df, "vmdr_scanners", cnxn, dtype=COLS, override_import_dt=override_import_dt
    )


def upload_vmdr_static_searchlists(
    searchlists: BaseList, cnxn: Connection, override_import_dt: datetime = None
) -> int:
    """
    Upload data from vmdr.get_static_searchlists() to SQL.

    Args:
        searchlists (BaseList): The Search List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "ID": types.Integer(),
        "TITLE": types.String(),
        "GLOBAL": types.Boolean(),
        "OWNER": types.String(),
        "CREATED": types.DateTime(),
        "MODIFIED": types.DateTime(),
        "MODIFIED_BY": types.String(),
        "QIDS": types.String(),  # BaseList
        "OPTION_PROFILES": types.String(),  # dict
        "REPORT_TEMPLATES": types.String(),  # dict
        "REMEDIATION_POLICIES": types.String(),  # dict
        "DISTRIBUTION_GROUPS": types.String(),  # dict
        "COMMENTS": types.String(),  # dict
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(searchlist) for searchlist in searchlists])

    # Upload the data:
    return upload_data(
        df,
        "vmdr_static_searchlists",
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_users(
    users: BaseList, cnxn: Connection, override_import_dt: datetime = None
) -> int:
    """
    Upload data from vmdr.get_user_list() to SQL.

    Args:
        users (BaseList): The user list to upload.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "USER_LOGIN": types.String(),
        "USER_ID": types.Integer(),
        "EXTERNAL_ID": types.String(),
        "FIRSTNAME": types.String(),
        "LASTNAME": types.String(),
        "TITLE": types.String(),
        "PHONE": types.String(),
        "FAX": types.String(),
        "EMAIL": types.String(),
        "COMPANY": types.String(),
        "ADDRESS1": types.String(),
        "ADDRESS2": types.String(),
        "CITY": types.String(),
        "COUNTRY": types.String(),
        "STATE": types.String(),
        "ZIP_CODE": types.String(),
        "TIME_ZONE_CODE": types.String(),
        "USER_STATUS": types.String(),
        "CREATION_DATE": types.DateTime(),
        "USER_ROLE": types.String(),
        "BUSINESS_UNIT": types.String(),
        "MANAGER_POC": types.String(),
        "UI_INTERFACE_STYLE": types.String(),
        "CREATE_OPTION_PROFILES": types.Boolean(),
        "PURGE_INFO": types.Boolean(),
        "ADD_ASSETS": types.Boolean(),
        "EDIT_REMEDIATION_POLICY": types.Boolean(),
        "EDIT_AUTH_RECORDS": types.Boolean(),
        "LATEST_VULN": types.String(),
        "MAP": types.String(),
        "DAILY_TICKETS": types.Boolean(),
        "ASSET_GROUP_TITLE": types.String(),  # BaseList
        "LAST_LOGIN_DATE": types.DateTime(),
        "UNIT_MANAGER_POC": types.Float(),
        "MAP": types.String(),
        "SCAN": types.String(),
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(user) for user in users])

    # Drop contact_info, assigned_asset_groups,
    # permissions, notifications:
    df.drop(
        columns=[
            "CONTACT_INFO",
            "ASSIGNED_ASSET_GROUPS",
            "PERMISSIONS",
            "NOTIFICATIONS",
        ],
        inplace=True,
    )

    # Upload the data:
    return upload_data(
        df,
        "vmdr_users",
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )


def upload_vmdr_scanlist(
    scans: BaseList, cnxn: Connection, override_import_dt: datetime = None
) -> int:
    """
    Upload data from vmdr.get_scan_list() to SQL.

    Args:
        scans (BaseList): The Scan List to upload.
        cnxn (Connection): The Connection object to the SQL database.
        override_import_dt (datetime): Use the passed datetime instead of generating one to upload to the database.

    Returns:
        int: The number of rows uploaded.
    """

    COLS = {
        "REF": types.String(),
        "TYPE": types.String(),
        "TITLE": types.String(),
        "USER_LOGIN": types.String(),
        "LAUNCH_DATETIME": types.DateTime(),
        "DURATION": types.String(),
        "PROCESSING_PRIORITY": types.String(),
        "PROCESSED": types.Boolean(),
        "STATE": types.String(),
        "TARGET": types.String(),  # BaseList
        "OPTION_PROFILE": types.String(),  # dict
        "ASSET_GROUP_TITLE_LIST": types.String(),  # BaseList
    }

    # Convert the BaseList to a DataFrame:
    df = DataFrame([prepare_dataclass(scan) for scan in scans])

    # Drop the STATUS column, as it is parsed out into STATE:
    df.drop(columns=["STATUS"], inplace=True)

    # Upload the data:
    return upload_data(
        df,
        "vmdr_scans",
        cnxn,
        dtype=COLS,
        override_import_dt=override_import_dt,
    )



"""
call_api.py - contains the call_api function for the qualysdk package.

This function handles all API calls to the Qualys API. It takes in a URL, headers, and a payload, and returns the response from the API.
Qualys uses many tricks in their API, such as using both url params and post data.
"""

from requests import request, Response
from typing import Literal, Union
from datetime import datetime

from ..auth.token import TokenAuth
from ..auth.basic import BasicAuth
from ..exceptions.Exceptions import *
from .call_schema import CALL_SCHEMA
from .convert_bools_and_nones import convert_bools_and_nones
from .xml_parser import xml_parser


def call_api(
    auth: Union[BasicAuth, TokenAuth],
    module: str,
    endpoint: str,
    headers: dict = None,
    params: dict = None,
    payload: dict = None,
    jsonbody: dict = None,
    override_method: Literal["GET", "POST", "PUT", "PATCH", "DELETE"] = None,
) -> Response:
    """
    Base call function for the Qualys API.

    This function is used across all modules and endpoints as the actual
    API call function.

    Params:
    ```
    auth (Union[BasicAuth, TokenAuth]) The authentication object.
    module (str) The module to call using the CALL_SCHEMA.
    endpoint (str) The endpoint to call using the CALL_SCHEMA.
    headers (dict) The headers to send.
    payload (dict) The payload to send.
    params (dict) The parameters to send.
    jsonbody (dict) The JSON body to send.
    override_method (Literal["GET", "POST", "PUT", "PATCH", "DELETE"]) The method to override the schema with.
    ```
    """

    # check module and endpoint:
    if module.lower() not in CALL_SCHEMA.keys():
        raise ValueError(
            f"Invalid module {module}. Valid modules are: {CALL_SCHEMA.keys()}."
        )
    if endpoint.lower() not in CALL_SCHEMA[module].keys():
        raise ValueError(
            f"Invalid endpoint {endpoint} for module {module}. Valid endpoints are: {[i for i in CALL_SCHEMA[module].keys() if i != 'url_type']}."
        )

    SCHEMA = CALL_SCHEMA[module][endpoint]

    # check the auth type:
    if SCHEMA["auth_type"] != auth.auth_type:
        raise AuthTypeMismatchError(
            f"Auth type mismatch. Expected {SCHEMA['auth_type']} but got {auth.auth_type}."
        )

    # check override:
    if override_method:
        if override_method.upper() not in SCHEMA["method"]:
            raise ValueError(
                f"Invalid override method {override_method}. Valid methods are: {SCHEMA['method']}."
            )

    # match the url_type to get the proper template:
    match CALL_SCHEMA[module]["url_type"]:
        case "gateway":
            url = f"https://gateway.{auth.platform}.apps.qualys.com{SCHEMA['endpoint']}"
        case "api":
            if (
                auth.platform == "qg1"
            ):  # Special case for qg1 platform: no qg or apps in the URL
                url = f"https://qualysapi.qualys.com{SCHEMA['endpoint']}"
            else:
                url = f"https://qualysapi.{auth.platform}.apps.qualys.com{SCHEMA['endpoint']}"
        case _:
            raise ValueError(f"Invalid url_type {SCHEMA['url_type']}.")

    # if token auth, check if token is not 4+ hours old:
    if isinstance(auth, TokenAuth):
        # check that the time delta between now and the token generation time is less than 4 hours:
        if (datetime.now() - auth.generated_on).seconds > 14400:
            print("Token is 4+ hours old. Refreshing token.")
            auth.get_token()

    # check params:
    if params:
        for key in params.keys():
            if key not in SCHEMA["valid_params"]:
                raise ValueError(
                    f"Invalid parameter {key} for {module}-{endpoint}. Valid parameters are: {SCHEMA['valid_params']}."
                )

    # check post data:
    if payload:
        for key in payload.keys():
            if key not in SCHEMA["valid_POST_data"]:
                raise ValueError(f"Invalid payload key {key} for {module}-{endpoint}.")

    # check if data should be POSTed as requests.post(json=):
    if SCHEMA["use_requests_json_data"]:
        use_json = True
    else:
        use_json = False

    # set up JWT auth header if needed:
    if auth.auth_type == "token":
        headers = auth.as_header()
    # or set up the tuple for basic auth:
    elif auth.auth_type == "basic":
        auth_tuple = (auth.username, auth.password)

    # Make certain payloads/params requests-friendly:
    if payload:
        payload = convert_bools_and_nones(payload)
    if params:
        params = convert_bools_and_nones(params)

    # and finally, make the request:
    response = request(
        method=SCHEMA["method"][0] if not override_method else override_method.upper(),
        url=url,
        headers=headers,
        params=params,
        data=payload if not use_json else None,
        json=jsonbody if use_json else None,
        auth=(auth_tuple if auth.auth_type == "basic" else None),
    )

    # check for errors:
    if response.status_code in range(400, 599):
        # print(f"Error: {response.text}")
        # response.raise_for_status()
        parsed = xml_parser(response.text)
        # Common path is [SIMPLE_RETURN][RESPONSE][TEXT]
        if "SIMPLE_RETURN" in parsed:
            raise QualysAPIError(parsed["SIMPLE_RETURN"]["RESPONSE"]["TEXT"])
        elif "html" in parsed:
            raise Exception(
                f"Error: {parsed['html']['body']['h1']}: {parsed['html']['body']['p'][1]['#text']}"
            )
        else:
            # yeah... turns out a homegrown xml parsers aren't super easy...
            raise Exception(
                f"Error: {parsed['{http://www.w3.org/1999/xhtml}html']['{http://www.w3.org/1999/xhtml}body']['{http://www.w3.org/1999/xhtml}h1']}"
            )

    return response



"""
get_host_list_detections.py - contains the get_host_list_detections function for the qualysdk package.

This endpoint is used to get a list of hosts and their QID detections. The function is multithreaded and uses the hld_backend function to pull the data.
"""

from typing import List, Union
from urllib.parse import parse_qs, urlparse
from threading import current_thread, Thread, Lock
from queue import Queue, Empty
from os import cpu_count

from .data_classes.hosts import VMDRHost, VMDRID
from .data_classes.lists.base_list import BaseList
from ..base.call_api import call_api
from ..auth.token import BasicAuth
from .get_host_list import (
    get_host_list,
)  # Used to grab list of IDs for multithreaded detection list pulls
from ..exceptions.Exceptions import *
from ..base.xml_parser import xml_parser

LOCK = Lock()


def normalize_id_list(id_list):
    """
    normalize_id_list - formats the kwarg ids, if it is passed.

    Since the API accepts multiple values, either as comma-separated or as a range, this function
    normalizes the input to a list of integers that satisfy the API requirements.
    """
    id_list = id_list.split(",")
    new_list = []
    for i in id_list:
        if "-" in i:
            # if there is a hyphen, split by hyphen and create a range
            new_list.extend(range(int(i.split("-")[0]), int(i.split("-")[1]) + 1))
        else:
            # if there is no hyphen, just append the integer
            new_list.append(int(i))
    # Sort the list to ensure the range is correct
    new_list.sort()
    return new_list


def pull_id_set(auth: BasicAuth, ids: str = None) -> BaseList[VMDRID]:
    """
    pull_id_set - pull a set of host IDs from the VMDR API.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        ids (str): A comma-separated string of host IDs to use. If specified, this will be used instead of pulling the full set.

    Returns:
        List[int]: A BaseList of host IDs as VMDRIDs.
    """
    if ids:
        return [
            str(i)
            for i in get_host_list(auth, details=None, truncation_limit=0, ids=ids)
        ]
    else:
        return [str(i) for i in get_host_list(auth, details=None, truncation_limit=0)]


def hld_backend(
    auth: BasicAuth,
    page_count: Union[int, "all"] = "all",
    **kwargs,
) -> List:
    """
    hld_backend - get a list of hosts and their QID detections.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        page_count (Union[int, "all"]): The number of pages to retrieve. Defaults to "all".
        **kwargs: Additional keyword arguments to pass to the API. See below.

    Kwargs:

        action (Optional[str]) #The action to perform. Default is 'list'. WARNING: any value you pass is overwritten with 'list'. It is just recognized as valid for the sake of completeness.
        echo_request (Optional[bool]) #Whether to echo the request. Default is False. Ends up being passed to the API as 0 or 1. WARNING: this SDK does not include this field in the data.
        show_asset_id (Optional[bool]) #Whether to show the asset IDs. Default is 'False'. ends up being passed to API as 0 or 1.
        include_vuln_type (Optional[Literal["confirmed", "potential"]]) #The type of vulnerability to include. If not specified, both types are included.

        DETECTION FILTERS:
        show_results (Optional[bool]) #Whether to show the results. Default is True. Ends up being passed to the API as 0 or 1. WARNING: this SDK overwrites any value you pass with '1'. It is just recognized as valid for the sake of completeness.
        arf_kernel_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns for Linux kernels. 0 = don't filter, 1 = exclude kernel vulns that are not exploitable, 2 = only include kernel related vulns that are not exploitable, 3 = only include exploitable kernel vulns, 4 = only include kernel vulns. If specified, results are in a host's <AFFECT_RUNNING_KERNEL> tag.
        arf_service_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns found on running or nonrunning ports/services. 0 = don't filter, 1 = exclude service related vulns that are exploitable, 2 = only include service vulns that are exploitable, 3 = only include service vulns that are not exploitable, 4 = only include service vulns. If specified, results are in a host's <AFFECT_RUNNING_SERVICE> tag.
        arf_config_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns that can be vulnerable based on host config. 0 = don't filter, 1 = exclude vulns that are exploitable due to host config, 2 = only include config related vulns that are exploitable, 3 = only include config related vulns that are not exploitable, 4 = only include config related vulns. If specified, results are in a host's <AFFECT_EXPLOITABLE_CONFIG> tag.
        active_kernels_only (Optional[Literal[0,1,2,3]]) #Specify vulns related to running or non-running kernels. 0 = don't filter, 1 = exclude non-running kernels, 2 = only include vulns on non-running kernels, 3 = only include vulns with running kernels. If specified, results are in a host's <AFFECT_RUNNING_KERNEL> tag.
        output_format (Optional[Literal["XML", "CSV"]]) #The format of the output. Default is 'XML'. WARNING: this SDK will overwrite any value you pass with 'XML'. It is just recognized as valid for the sake of completeness.
        supress_duplicated_data_from_csv (Optional[bool]) #Whether to suppress duplicated data from CSV. Default is False. Ends up being passed to the API as 0 or 1. WARNING: this SDK does not include this field in the data.
        truncation_limit (Optional[int]) #The truncation limit for a page. Default is 1000.
        max_days_since_detection_updated (Optional[int]) #The maximum number of days since the detection was last updated. For detections that have never changed, the value is applied to the last detection date.
        detection_updated_since (Optional[str]) #The date and time since the detection was updated.
        detection_updated_before (Optional[str]) #The date and time before the detection was updated.
        detection_processed_before (Optional[str]) #The date and time before the detection was processed.
        detection_processed_after (Optional[str]) #The date and time after the detection was processed.
        detection_last_tested_since (Optional[str]) #The date and time since the detection was last tested.
        detection_last_tested_since_days (Optional[int]) #The number of days since the detection was last tested.
        detection_last_tested_before (Optional[str]) #The date and time before the detection was last tested.
        detection_last_tested_before_days (Optional[int]) #The number of days before the detection was last tested.
        include_ignored (Optional[bool]) #Whether to include ignored detections. Default is False. Ends up being passed to the API as 0 or 1.
        include_disabled (Optional[bool]) #Whether to include disabled detections. Default is False. Ends up being passed to the API as 0 or 1.

        HOST FILTERS:
        ids (Optional[str]) #A comma-separated string of host IDs to include.
        id_min (Optional[Union[int,str]]) #The minimum host ID to include.
        id_max (Optional[Union[int,str]]) #The maximum host ID to include.
        ips (Optional[str]) #The IP address of the host to include. Can be a comma-separated string, and also supports ranges with a hyphen: 10.0.0.0-10.0.0.255.
        ipv6 (Optional[str]) #The IPv6 address of the host to include. Can be a comma-separated string. Does not support ranges.
        ag_ids (Optional[str]) #Show only hosts belonging to the specified asset group IDs. Can be a comma-separated string, and also supports ranges with a hyphen: 1-5.
        ag_titles (Optional[str]) #Show only hosts belonging to the specified asset group titles. Can be a comma-separated string.
        network_ids (Optional[str]) #Show only hosts belonging to the specified network IDs. Can be a comma-separated string.
        network_names (Optional[str]) #displays the name of the network corresponding to the network ID.
        vm_scan_since (Optional[str]) #The date and time since the last VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        no_vm_scan_since (Optional[str]) #The date and time since the last VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        max_days_since_last_vm_scan (Optional[int]) #The maximum number of days since the last VM scan.
        vm_processed_before (Optional[str]) #The date and time before the VM scan was processed. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_processed_after (Optional[str]) #The date and time after the VM scan was processed. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_scan_date_before (Optional[str]) #The date and time before the VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_scan_date_after (Optional[str]) #The date and time after the VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_auth_scan_date_before (Optional[str]) #The date and time before the VM authenticated scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_auth_scan_date_after (Optional[str]) #The date and time after the VM authenticated scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        status (Optional[Literal["New", "Active", "Fixed", "Re-Opened"]]) #The status of the detection.
        compliance_enabled (Optional[bool]) #Whether compliance is enabled. Default is False. Ends up being passed to the API as 0 or 1.
        os_pattern (Optional[str]) #PCRE Regex to match operating systems.

        QID FILTERS:
        qids (Optional[str]) #A comma-separated string of QIDs to include.
        severities (Optional[str]) #A comma-separated string of severities to include. Can also be a hyphenated range, i.e. '2-4'.
        filter_superseded_qids (Optional[bool]) #Whether to filter superseded QIDs. Default is False. Ends up being passed to the API as 0 or 1.
        include_search_list_titles (Optional[str]) #A comma-separated string of search list titles to include.
        exclude_search_list_titles (Optional[str]) #A comma-separated string of search list titles to exclude.
        include_search_list_ids (Optional[str]) #A comma-separated string of search list IDs to include.
        exclude_search_list_ids (Optional[str]) #A comma-separated string of search list IDs to exclude.

        ASSET TAG FILTERS:
        use_tags (Optional[bool]) #Whether to use tags. Default is False. Ends up being passed to the API as 0 or 1.
        tag_set_by (Optional[Literal['id','name']]) #When filtering on tags, whether to filter by tag ID or tag name.
        tag_include_selector (Optional[Literal['any','all']]) #When filtering on tags, choose if asset has to match any or all tags specified.
        tag_exclude_selector (Optional[Literal['any','all']]) #When filtering on tags, choose if asset has to match any or all tags specified.
        tag_set_include (Optional[str]) #A comma-separated string of tag IDs or names to include.
        tag_set_exclude (Optional[str]) #A comma-separated string of tag IDs or names to exclude.
        show_tags (Optional[bool]) #Whether to show tags. Default is False. Ends up being passed to the API as 0 or 1.

        QDS FILTERS:
        show_qds (Optional[bool]) #Whether to show QDS. Default is False. Ends up being passed to the API as 0 or 1.
        qds_min (Optional[int]) #The minimum QDS to include.
        qds_max (Optional[int]) #The maximum QDS to include.
        show_qds_factors (Optional[bool]) #Whether to show QDS factors. Default is False. Ends up being passed to the API as 0 or 1.

        EC2/AZURE METADATA FILTERS:
        host_metadata (Optional[Literal['all,'ec2', 'azure']]) #The type of metadata to include. Default is 'all'.
        host_metadata_fields (Optional[str]) #A comma-separated string of metadata fields to include. Use carefully.
        show_cloud_tags (Optional[bool]) #Whether to show cloud tags. Default is False. Ends up being passed to the API as 0 or 1.
        cloud_tag_fields (Optional[str]) #A comma-separated string of cloud tag fields to include. Use carefully.

    Returns:
        List: A list of VMDRHost objects, with their DETECTIONS attribute populated.
    """

    # Set the kwargs
    kwargs["action"] = "list"
    kwargs["echo_request"] = 0
    kwargs["show_results"] = 1
    kwargs["output_format"] = "XML"

    responses = BaseList()
    pulled = 0

    while True:
        with LOCK:
            print(
                f"{current_thread().name} - Pulling page {pulled+1} for ids {kwargs.get('ids')}. KWARGS: {kwargs}"
            )

        # make the request:
        response = call_api(
            auth=auth,
            module="vmdr",
            endpoint="get_hld",
            params=kwargs,
            headers={"X-Requested-With": "qualysdk SDK"},
        )

        if response.status_code != 200:
            with LOCK:
                print(f"{current_thread().name} - No data returned on page {pulled}")
            pulled += 1
            if pulled != "all":
                if pulled == page_count:
                    print(f"{current_thread().name} - Pulled all pages.")
                    break
                else:
                    continue

        # cleaned = remove_problem_characters(response.text)
        xml = xml_parser(response.content)

        # check if there is no host list
        if "HOST_LIST" not in xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]:
            with LOCK:
                print(f"{current_thread().name} - No host list returned.")
        else:
            # check if ["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"] is a list of dictionaries
            # or just a dictionary. if it is just one, put it inside a list
            if not isinstance(
                xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"]["HOST"],
                list,
            ):
                xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"][
                    "HOST"
                ] = [
                    xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"][
                        "HOST"
                    ]
                ]

            for host in xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["HOST_LIST"][
                "HOST"
            ]:
                host_obj = VMDRHost.from_dict(host)
                responses.append(host_obj)

        pulled += 1
        if page_count != "all":
            if pulled == page_count:
                break

        if "WARNING" in xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]:
            if "URL" in xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["WARNING"]:
                # get the id_min parameter from the URL to pass into kwargs:
                params = parse_qs(
                    urlparse(
                        xml["HOST_LIST_VM_DETECTION_OUTPUT"]["RESPONSE"]["WARNING"][
                            "URL"
                        ]
                    ).query
                )
                with LOCK:
                    print(
                        f"{current_thread().name} - Pagination detected. Pulling next page with id_min: {params['id_min'][0]}"
                    )
                kwargs["id_min"] = params["id_min"][0]

            else:
                break
        else:
            break

    return responses


def create_id_queue(
    auth: BasicAuth, chunk_size: int = 100, ids: str = None
) -> BaseList:
    """
    create_id_queue - create a queue of host IDs to pull. the queue contains chunks (lists) of chunk_size
    length with host IDs.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        chunk_size (int): The size of each chunk. Defaults to 100.
        ids (str): A comma-separated string of host IDs to use. If specified, this will be used instead of pulling the full set.

    Returns:
        Queue: A queue of host IDs to pull.
    """

    if ids:
        id_list = pull_id_set(auth, ids=ids)
    else:
        id_list = pull_id_set(auth)

    if not id_list:
        raise QualysAPIError("No IDs returned from API.")

    print(f"ID set pulled. Total IDs: {len(id_list)}")

    id_queue = Queue()

    for i in range(0, len(id_list), chunk_size):
        id_queue.put(id_list[i : i + chunk_size])

    singular_chunk = True if id_queue.qsize() == 1 else False

    print(
        f"Queue created with {id_queue.qsize()} {'chunks' if not singular_chunk else 'chunk'} of ~{chunk_size} IDs{' each.' if not singular_chunk else '.'}"
    )

    return id_queue


def get_hld(
    auth: BasicAuth,
    chunk_size: int = 3000,
    threads: int = 5,
    page_count: Union[int, "all"] = "all",
    chunk_count: Union[int, "all"] = "all",
    **kwargs,
) -> List:
    """
    get_hld - get a list of hosts and their QID detections using multiple threads. Read
    hld_backend for more information on the kwargs.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        chunk_size (int): The size of each chunk. Defaults to 3000.
        threads (int): The number of threads to use. Defaults to 5.
        page_count (Union[int, "all"]): The number of pages to retrieve. Defaults to "all".
        chunk_count (Union[int, "all"]): The number of chunks to retrieve. Defaults to "all".
        **kwargs: Additional keyword arguments to pass to the API.

    Kwargs:

        action (Optional[str]) #The action to perform. Default is 'list'. WARNING: any value you pass is overwritten with 'list'. It is just recognized as valid for the sake of completeness.
        echo_request (Optional[bool]) #Whether to echo the request. Default is False. Ends up being passed to the API as 0 or 1. WARNING: this SDK does not include this field in the data.
        show_asset_id (Optional[bool]) #Whether to show the asset IDs. Default is 'False'. ends up being passed to API as 0 or 1.
        include_vuln_type (Optional[Literal["confirmed", "potential"]]) #The type of vulnerability to include. If not specified, both types are included.

        DETECTION FILTERS:
        show_results (Optional[bool]) #Whether to show the results. Default is True. Ends up being passed to the API as 0 or 1. WARNING: this SDK overwrites any value you pass with '1'. It is just recognized as valid for the sake of completeness.
        arf_kernel_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns for Linux kernels. 0 = don't filter, 1 = exclude kernel vulns that are not exploitable, 2 = only include kernel related vulns that are not exploitable, 3 = only include exploitable kernel vulns, 4 = only include kernel vulns. If specified, results are in a host's <AFFECT_RUNNING_KERNEL> tag.
        arf_service_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns found on running or nonrunning ports/services. 0 = don't filter, 1 = exclude service related vulns that are exploitable, 2 = only include service vulns that are exploitable, 3 = only include service vulns that are not exploitable, 4 = only include service vulns. If specified, results are in a host's <AFFECT_RUNNING_SERVICE> tag.
        arf_config_filter (Optional[Literal[0,1,2,3,4]]) #Specify vulns that can be vulnerable based on host config. 0 = don't filter, 1 = exclude vulns that are exploitable due to host config, 2 = only include config related vulns that are exploitable, 3 = only include config related vulns that are not exploitable, 4 = only include config related vulns. If specified, results are in a host's <AFFECT_EXPLOITABLE_CONFIG> tag.
        active_kernels_only (Optional[Literal[0,1,2,3]]) #Specify vulns related to running or non-running kernels. 0 = don't filter, 1 = exclude non-running kernels, 2 = only include vulns on non-running kernels, 3 = only include vulns with running kernels. If specified, results are in a host's <AFFECT_RUNNING_KERNEL> tag.
        output_format (Optional[Literal["XML", "CSV"]]) #The format of the output. Default is 'XML'. WARNING: this SDK will overwrite any value you pass with 'XML'. It is just recognized as valid for the sake of completeness.
        supress_duplicated_data_from_csv (Optional[bool]) #Whether to suppress duplicated data from CSV. Default is False. Ends up being passed to the API as 0 or 1. WARNING: this SDK does not include this field in the data.
        truncation_limit (Optional[int]) #The truncation limit for a page. Default is 1000.
        max_days_since_detection_updated (Optional[int]) #The maximum number of days since the detection was last updated. For detections that have never changed, the value is applied to the last detection date.
        detection_updated_since (Optional[str]) #The date and time since the detection was updated.
        detection_updated_before (Optional[str]) #The date and time before the detection was updated.
        detection_processed_before (Optional[str]) #The date and time before the detection was processed.
        detection_processed_after (Optional[str]) #The date and time after the detection was processed.
        detection_last_tested_since (Optional[str]) #The date and time since the detection was last tested.
        detection_last_tested_since_days (Optional[int]) #The number of days since the detection was last tested.
        detection_last_tested_before (Optional[str]) #The date and time before the detection was last tested.
        detection_last_tested_before_days (Optional[int]) #The number of days before the detection was last tested.
        include_ignored (Optional[bool]) #Whether to include ignored detections. Default is False. Ends up being passed to the API as 0 or 1.
        include_disabled (Optional[bool]) #Whether to include disabled detections. Default is False. Ends up being passed to the API as 0 or 1.

        HOST FILTERS:
        ids (Optional[str]) #A comma-separated string of host IDs to include.
        id_min (Optional[Union[int,str]]) #The minimum host ID to include.
        id_max (Optional[Union[int,str]]) #The maximum host ID to include.
        ips (Optional[str]) #The IP address of the host to include. Can be a comma-separated string, and also supports ranges with a hyphen: 10.0.0.0-10.0.0.255.
        ipv6 (Optional[str]) #The IPv6 address of the host to include. Can be a comma-separated string. Does not support ranges.
        ag_ids (Optional[str]) #Show only hosts belonging to the specified asset group IDs. Can be a comma-separated string, and also supports ranges with a hyphen: 1-5.
        ag_titles (Optional[str]) #Show only hosts belonging to the specified asset group titles. Can be a comma-separated string.
        network_ids (Optional[str]) #Show only hosts belonging to the specified network IDs. Can be a comma-separated string.
        network_names (Optional[str]) #displays the name of the network corresponding to the network ID.
        vm_scan_since (Optional[str]) #The date and time since the last VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        no_vm_scan_since (Optional[str]) #The date and time since the last VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        max_days_since_last_vm_scan (Optional[int]) #The maximum number of days since the last VM scan.
        vm_processed_before (Optional[str]) #The date and time before the VM scan was processed. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_processed_after (Optional[str]) #The date and time after the VM scan was processed. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_scan_date_before (Optional[str]) #The date and time before the VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_scan_date_after (Optional[str]) #The date and time after the VM scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_auth_scan_date_before (Optional[str]) #The date and time before the VM authenticated scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        vm_auth_scan_date_after (Optional[str]) #The date and time after the VM authenticated scan. Format is 'YYYY-MM-DD[THH:MM:SS]'.
        status (Optional[Literal["New", "Active", "Fixed", "Re-Opened"]]) #The status of the detection.
        compliance_enabled (Optional[bool]) #Whether compliance is enabled. Default is False. Ends up being passed to the API as 0 or 1.
        os_pattern (Optional[str]) #PCRE Regex to match operating systems.

        QID FILTERS:
        qids (Optional[str]) #A comma-separated string of QIDs to include.
        severities (Optional[str]) #A comma-separated string of severities to include. Can also be a hyphenated range, i.e. '2-4'.
        filter_superseded_qids (Optional[bool]) #Whether to filter superseded QIDs. Default is False. Ends up being passed to the API as 0 or 1.
        include_search_list_titles (Optional[str]) #A comma-separated string of search list titles to include.
        exclude_search_list_titles (Optional[str]) #A comma-separated string of search list titles to exclude.
        include_search_list_ids (Optional[str]) #A comma-separated string of search list IDs to include.
        exclude_search_list_ids (Optional[str]) #A comma-separated string of search list IDs to exclude.

        ASSET TAG FILTERS:
        use_tags (Optional[bool]) #Whether to use tags. Default is False. Ends up being passed to the API as 0 or 1.
        tag_set_by (Optional[Literal['id','name']]) #When filtering on tags, whether to filter by tag ID or tag name.
        tag_include_selector (Optional[Literal['any','all']]) #When filtering on tags, choose if asset has to match any or all tags specified.
        tag_exclude_selector (Optional[Literal['any','all']]) #When filtering on tags, choose if asset has to match any or all tags specified.
        tag_set_include (Optional[str]) #A comma-separated string of tag IDs or names to include.
        tag_set_exclude (Optional[str]) #A comma-separated string of tag IDs or names to exclude.
        show_tags (Optional[bool]) #Whether to show tags. Default is False. Ends up being passed to the API as 0 or 1.

        QDS FILTERS:
        show_qds (Optional[bool]) #Whether to show QDS. Default is False. Ends up being passed to the API as 0 or 1.
        qds_min (Optional[int]) #The minimum QDS to include.
        qds_max (Optional[int]) #The maximum QDS to include.
        show_qds_factors (Optional[bool]) #Whether to show QDS factors. Default is False. Ends up being passed to the API as 0 or 1.

        EC2/AZURE METADATA FILTERS:
        host_metadata (Optional[Literal['all,'ec2', 'azure']]) #The type of metadata to include. Default is 'all'.
        host_metadata_fields (Optional[str]) #A comma-separated string of metadata fields to include. Use carefully.
        show_cloud_tags (Optional[bool]) #Whether to show cloud tags. Default is False. Ends up being passed to the API as 0 or 1.
        cloud_tag_fields (Optional[str]) #A comma-separated string of cloud tag fields to include. Use carefully.

    Returns:
        BaseList: A list of VMDRHost objects, with their DETECTIONS attribute populated.
    """

    # Ensure that threads, chunk_size, and page_count (if not 'all') are all integers above 0
    if any(
        [
            threads < 1,
            chunk_size < 1,
            (page_count != "all" and page_count < 1),
            (chunk_count != "all" and chunk_count < 1),
        ]
    ):
        raise ValueError(
            "threads, chunk_size, page_count (if not 'all') and chunk_count (if not 'all') must all be integers above 0."
        )

    # Make sure the user hasn't set threads to more than the cpu count
    if threads > cpu_count():
        print(
            f"Warning: The number of threads ({threads}) is greater than the number of CPUs ({cpu_count()}). This may cause performance issues."
        )

    # Second, get concurrency rate limit from auth object. NOTE: eventually, auth class should have an attribute for this.
    rl = auth.get_ratelimit()
    if threads > rl["X-Concurrency-Limit-Limit"]:
        print(
            f"Warning: The number of threads ({threads}) is greater than the concurrency rate limit ({rl['X-Concurrency-Limit-Limit']}). Setting threads to {rl['X-Concurrency-Limit-Limit']}."
        )
        threads = rl["X-Concurrency-Limit-Limit"]

    (
        print(f"Pulling/creating queue for full ID list...")
        if not kwargs.get("ids")
        else print(
            f"Pulling/creating queue for user-specified IDs: {kwargs.get('ids')}..."
        )
    )

    id_queue = create_id_queue(auth, chunk_size=chunk_size, ids=kwargs.get("ids"))
    print(f"Starting get_hld with {threads} {'threads.' if threads > 1 else 'thread.'}")

    threads_list = []

    responses = BaseList()

    for i in range(threads):
        thread = Thread(
            target=threaded_hld_worker,
            args=(auth, id_queue, responses, page_count, chunk_count, kwargs),
        )
        threads_list.append(thread)
        thread.start()

    for thread in threads_list:
        thread.join()

    print("All threads have completed. Returning responses.")
    return responses


def threaded_hld_worker(
    auth: BasicAuth,
    id_queue: Queue,
    responses: BaseList,
    page_count: Union[int, "all"],
    chunk_count: Union[int, "all"],
    kwargs,
):
    """
    threaded_hld_worker - the worker function for get_hld/hld_backend functions.

    Params:
        auth (BasicAuth): The BasicAuth object containing the username and password.
        id_queue (Queue): The queue of host IDs to pull.
        responses (BaseList): The list of responses to append to.
        page_count (Union[int, "all"]): The number of pages to retrieve. Defaults to "all".
        chunk_count (Union[int, "all"]): The number of chunks to retrieve. Defaults to "all".
        **kwargs: Additional keyword arguments to pass to the API. See get_hld() for details.
    """
    while True:
        pages_pulled = 0
        chunks_pulled = 0
        try:
            ids = (
                id_queue.get_nowait()
            )  # nowait allows us to check if the queue is empty without blocking
        except Empty:
            with LOCK:
                print(f"{current_thread().name} - Queue is empty. Terminating thread.")
            break

        if not ids:
            with LOCK:
                print(f"{current_thread().name} - No IDs to pull. Terminating thread.")
            break

        kwargs["ids"] = f"{ids[0]}-{ids[-1]}"
        responses.extend(hld_backend(auth, page_count=page_count, **kwargs))
        id_queue.task_done()
        with LOCK:
            print(f"{current_thread().name} - Chunk complete.")
        pages_pulled += 1
        chunks_pulled += 1
        # check if the queue is empty, or if the threads are done (via pulled var)
        if id_queue.empty():
            with LOCK:
                print(f"{current_thread().name} - Queue is empty. Terminating thread.")
            break
        if pages_pulled == page_count:
            with LOCK:
                print(
                    f"{current_thread().name} - Thread has pulled all pages. Terminating thread."
                )
            break
        if chunks_pulled == chunk_count:
            with LOCK:
                print(
                    f"{current_thread().name} - Thread has pulled all chunks. Terminating thread."
                )
            break



"""
static_searchlist.py - Contains the search list dataclasses for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *
from datetime import datetime

from ..data_classes.lists import BaseList
from .kb_entry import KBEntry


@dataclass
class StaticSearchList:
    """
    StaticSearchList - represents a single Static Search List in VMDR.
    """

    ID: int = field(
        metadata={"description": "The ID of the Static Search List."}, default=None
    )
    TITLE: str = field(
        metadata={"description": "The title of the Static Search List."}, default=None
    )
    GLOBAL: bool = field(
        metadata={"description": "Whether the Static Search List is global."},
        default=None,
    )
    OWNER: str = field(
        metadata={"description": "The owner of the Static Search List."}, default=None
    )
    CREATED: Union[str, datetime] = field(
        metadata={"description": "The date the Static Search List was created."},
        default=None,
    )
    MODIFIED: Union[str, datetime] = field(
        metadata={"description": "The date the Static Search List was last modified."},
        default=None,
    )
    MODIFIED_BY: str = field(
        metadata={"description": "The user who last modified the Static Search List."},
        default=None,
    )
    QIDS: BaseList[KBEntry] = field(
        metadata={"description": "The QIDs associated with the Static Search List."},
        default=None,
    )

    OPTION_PROFILES: BaseList[dict] = field(
        metadata={
            "description": "The option profiles associated with the Static Search List."
        },
        default=None,
    )

    REPORT_TEMPLATES: BaseList[dict] = field(
        metadata={
            "description": "The report templates associated with the Static Search List."
        },
        default=None,
    )

    REMEDIATION_POLICIES: BaseList[dict] = field(
        metadata={
            "description": "The remediation policies associated with the Static Search List."
        },
        default=None,
    )

    DISTRIBUTION_GROUPS: BaseList[dict] = field(
        metadata={
            "description": "The distribution groups associated with the Static Search List."
        },
        default=None,
    )

    COMMENTS: str = field(
        metadata={
            "description": "The comments associated with the Static Search List."
        },
        default=None,
    )

    def __post_init__(self):
        # self.OPTION_PROFILES = BaseList(self.OPTION_PROFILES, dict)
        # self.REPORT_TEMPLATES = BaseList(self.REPORT_TEMPLATES, dict)
        # self.REMEDIATION_POLICIES = BaseList(self.REMEDIATION_POLICIES, dict)
        # self.DISTRIBUTION_GROUPS = BaseList(self.DISTRIBUTION_GROUPS, dict)

        if self.QIDS:
            bl = BaseList()
            # check for one QID
            if isinstance(self.QIDS["QID"], str):
                self.QIDS = [self.QIDS["QID"]]
            else:
                # Raise ['QID'] list up a level
                self.QIDS = self.QIDS["QID"]

            for qid in self.QIDS:
                bl.append(KBEntry(QID=qid))
            self.QIDS = bl

        DT_FIELDS = ["CREATED", "MODIFIED"]

        for field in DT_FIELDS:
            if getattr(self, field) and not isinstance(getattr(self, field), datetime):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        self.ID = int(self.ID)

        self.GLOBAL = True if self.GLOBAL != "No" else False

    def __str__(self):
        return self.TITLE

    def __dict__(self):
        return asdict(self)

    def __iter__(self):
        for key in self.__dict__():
            yield key, getattr(self, key)

    def keys(self):
        return self.__dict__().keys()

    def values(self):
        return self.__dict__().values()

    def items(self):
        return self.__dict__().items()

    def to_dict(self):
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict):
        return cls(**data)

    def contains_qid(self, qid: int) -> bool:
        """
        contains_qid - Determines if the Static Search List contains a specific QID.

        :param qid: The QID to check for.
        :type qid: int
        :return: Whether the QID is in the Static Search List.
        :rtype: bool
        """
        return any(detection.QID == qid for detection in self.QIDS)



"""
report_template.py - Contains the ReportTemplate data class.
"""

from typing import Union
from dataclasses import dataclass, field, asdict
from datetime import datetime


@dataclass
class ReportTemplate:
    """
    A single report template in VMDR
    """

    ID: int = field(
        metadata={"description": "The unique ID of the report template."}, default=None
    )

    TYPE: str = field(
        metadata={"description": "The type of the report template."}, default=None
    )

    TEMPLATE_TYPE: str = field(
        metadata={"description": "The template type of the report template."},
        default=None,
    )

    TITLE: str = field(
        metadata={"description": "The title of the report template."}, default=None
    )

    USER: dict = field(
        metadata={"description": "The user that created the report template."},
        default=None,
    )

    LAST_UPDATE: Union[str, datetime] = field(
        metadata={"description": "The last time the report template was updated."},
        default=None,
    )

    GLOBAL: bool = field(
        metadata={"description": "If the report template is global."}, default=None
    )

    def __post_init__(self):
        if self.LAST_UPDATE:
            self.LAST_UPDATE = datetime.fromisoformat(self.LAST_UPDATE)

        self.ID = int(self.ID)
        self.GLOBAL = bool(self.GLOBAL)

    @classmethod
    def from_dict(cls, data: dict):
        """
        Create a ReportTemplate object from a dictionary.
        """
        return cls(**data)

    def __dict__(self):
        return asdict(self)

    def to_dict(self):
        return asdict(self)

    def __int__(self):
        return self.ID

    def __str__(self):
        return f"{self.ID}: {self.TITLE}"

    def keys(self):
        return self.__dict__().keys()

    def values(self):
        return self.__dict__().values()

    def items(self):
        return self.__dict__().items()

    def valid_values(self):
        return {k: v for k, v in self.__dict__().items() if v}



"""
vmscans.py - Contains functions to interact with the /api/2.0/fo/scan/ API endpoint,
with the 'action' parameter controlling what is done with the scan(s).
"""

from datetime import datetime
from typing import Union, Literal
from io import StringIO

from pandas import read_json, DataFrame

from ..base import call_api, xml_parser
from .data_classes.lists.base_list import BaseList
from .data_classes.vmscan import VMScan
from ..auth.token import BasicAuth
from ..exceptions.Exceptions import *


def get_scan_list(auth: BasicAuth, **kwargs) -> BaseList[VMScan]:
    """
    Pull a list of scans in the Qualys subscription.

    Args:
        auth (BasicAuth): Qualys BasicAuth object.

    Keyword Args:
        SCAN LIST FILTERS:
        scan_ref (str): Scan reference ID. Formatted like scan/1234567890123456, compliance/1234567890123456, or qscap/1234567890123456.
        state (Union[Literal["Running", "Paused", "Canceled", "Finished", "Error", "Queued", "Loading"], BaseList[str]): State of the scan. Multiple states can be specified as a comma-separated string. Can also pass a BaseList of strings.
        processed (bool): Whether the scan has been processed. Defaults to None.
        type (Literal["On-Demand", "API", "Scheduled"]): Type of scan. Defaults to None.
        target (Union[str, BaseList[str]]) Target IP(s) of the scan. Can be a single IP string, or a BaseList of strings. Calls API with the correct format - a comma separated string. If an IP range/network obj is passed, it is formatted as 1.2.3.4-5.6.7.8.
        user_login (str): Filter on owner of the scan. Defaults to None.
        launched_after_datetime (Union[strm datetime]): Filter on scans launched after a certain datetime. Can be a string or a datetime object. If a datetime object is passed, it is converted to a string.
        launched_before_datetime (Union[str, datetime]): Filter on scans launched before a certain datetime. Can be a string or a datetime object. If a datetime object is passed, it is converted to a string.
        scan_type (Literal["certview", "ec2certview"]): Filter on scan type. Defaults to None.
        client_id (Union[str, int]): Filter on client ID. Defaults to None.
        client_name (str): Filter on client name. Defaults to None.

        SHOW/HIDE FIELDS:
        show_ags (bool): Whether to show AGs. Defaults to None.
        show_op (bool): Whether to show OP. Defaults to None.
        show_status (bool): Whether to show status. Defaults to None.
        show_last (bool): Whether to show last scan. Defaults to None.
        ignore_target (bool): Whether to ignore the target. Defaults to None.

    Returns:
        BaseList: BaseList object containing the scan(s) in the Qualys subscription.
    """

    DT_FIELDS = ["launched_after_datetime", "launched_before_datetime"]

    # If user passes a datetime object, convert it to a string:
    for field in DT_FIELDS:
        if field in kwargs and isinstance(kwargs[field], datetime):
            kwargs[field] = kwargs[field].strftime("%Y-%m-%dT%H:%M:%S")

    # For all other kwargs, check if they are BaseList/list objects and convert them to a comma-separated string:
    for key in kwargs:
        if type(kwargs[key]) in [BaseList, list]:
            kwargs[key] = ",".join(kwargs[key])

    kwargs["action"] = "list"

    # Make the request:
    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="list_scans",
        params=kwargs,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    # Parse the response:
    result = xml_parser(response.text)

    # Check for empty results:
    if not result:
        print("No scans found.")
        return None

    data = result["SCAN_LIST_OUTPUT"]["RESPONSE"]

    if "SCAN_LIST" not in data or "SCAN" not in data["SCAN_LIST"]:
        print("No scans found.")
        return None

    # If data["SCAN_LIST"]["SCAN"] is a dict, convert it to a list of dicts:
    if isinstance(data["SCAN_LIST"]["SCAN"], dict):
        data["SCAN_LIST"]["SCAN"] = [data["SCAN_LIST"]["SCAN"]]

    # Convert the data to a BaseList of VMScan objects:
    scans = BaseList()

    for scan in data["SCAN_LIST"]["SCAN"]:
        scans.append(VMScan.from_dict(scan))

    return scans


def launch_scan(auth: BasicAuth, **kwargs) -> VMScan:
    """
    Create a new VMDR scan and launch, or call
    a pre-existing VMDR scan on supplied target (IP, asset group IDs, etc.).

    Keyword args:
        action (Literal["launch"]): The action to perform. Defaults to "launch". WARNING: the SDK will force-set this to "launch".
        echo_request (bool): Whether to echo the request. Defaults to False. WARNING: the SDK will force-set this to False.
        runtime_http_header (str): The value the scanner will put in the Qualys-Scan header. Defaults to None. Used to "drop defenses".
        scan_title (str): The title of the scan. Defaults to None.
        option_id (int): The option profile ID. Required if option_title not specified.
        option_title (str): The option profile title. Required if option_id not specified.

        SCANNER APPLIANCE OPTIONS:
            iscanner_id (int): The internal scanner ID. Defaults to None.
            iscanner_name (str): The internal scanner name. Defaults to None.
            ec2_instance_ids (str): The EC2 instance IDs of external scanners. Defaults to None.

        priority (int, 0-10): The priority of the scan. Defaults to None.

        ASSET IPs/GROUP OPTIONS:
            ip (Union[str, BaseList[str]]): The IP(s) to scan. Defaults to None.
            asset_group_ids (Union[str, BaseList[str], int, BaseList[int]]): The asset group IDs to scan. Defaults to None.
            asset_groups (Union[str, BaseList[str]]): The asset group titles to scan. Defaults to None.
            exclude_ip_per_scan (Union[str, BaseList[str]]): The IPs to exclude from the scan. Defaults to None. Only valid when target_from=assets.
            fqdn (Union[str, BaseList[str]]): The FQDN(s) to scan. Defaults to None.
            default_scanner (bool): Whether to use the default scanner. Defaults to None.
            scanners_in_ag (bool): Whether to use scanners in the asset group. Defaults to None.
            target_from (Literal["assets", "tags"]): The target source. Defaults to "assets".
            use_ip_nt_range_tags_include (bool): Whether to use IP/NT range tags include. Defaults to None.
            use_ip_nt_range_tags_exclude (bool): Whether to use IP/NT range tags exclude. Defaults to None.
            use_ip_nt_range_tags (bool): Whether to use IP/NT range tags. Defaults to None.
            tag_include_selector (Literal["any", "all"]): The tag include selector. Defaults to any.
            tag_exclude_selector (Literal["any", "all"]): The tag exclude selector. Defaults to any.
            tag_set_by (Literal["id", "name"]) Whether to search for tags by ID or name. Defaults to id.
            tag_set_exclude (str): Comma-separated string of tag IDs or names, according to tag_set_by. Defaults to None.
            tag_set_include (str): Comma-separated string of tag IDs or names, according to tag_set_by. Defaults to None.

        ip_network_id (int): The IP network ID. Defaults to None. Must be enabled in the Qualys subscription.
        client_id (int): The client ID. Defaults to None. Only available for consultant subscriptions.
        client_name (str): The client name. Defaults to None. Only available for consultant subscriptions.
        connector_name (str): The connector name for EC2 scans. Defaults to None. Required for EC2 scans.
        ec2_endpoint (str): The EC2 region code or VPC ID zone. Defaults to None. Required for EC2 scans.

        EXAMPLE:
        ```
        # Launch a scan on a single IP:
        result = launch_scan(auth, ip="1.2.3.4", scan_title="My Scan", priority=5, option_id=123456)
        >>> "New vm scan launched with REF: scan/1234567890123456"
        result
        >>> ("New vm scan launched", "scan/1234567890123456")
        ```

    Returns:
        VMSan: VMScan dataclass object containing the scan details.
    """

    # Set the required kwargs:
    kwargs["action"] = "launch"
    kwargs["echo_request"] = False

    # Convert any BaseList objects to comma-separated strings:
    for key in kwargs:
        if isinstance(kwargs[key], BaseList):
            kwargs[key] = ",".join(kwargs[key])

    # Make the request:
    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="launch_scan",
        payload=kwargs,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    # Parse the response:
    result = xml_parser(response.text)

    # Check for empty results:
    if not result:
        print("No scan launched.")
        return None

    # Check for scan details in simple_return:

    data = result["SIMPLE_RETURN"]["RESPONSE"]

    if "CODE" in data.keys():
        raise QualysAPIError(data["TEXT"])

    items_list = data["ITEM_LIST"]["ITEM"]
    # if items_list is a dict, put it into a list:
    if isinstance(items_list, dict):
        items_list = [items_list]

    scan_ref = ""
    for item in items_list:
        # Check if the KEY key's value is REFERENCE
        if item["KEY"] == "REFERENCE":
            scan_ref = item["VALUE"]

    print(f'{data["TEXT"]} with REF: {scan_ref}')

    # Return a VMScan object with the scan details:
    return get_scan_list(auth, scan_ref=scan_ref)[0]


def manage_scan(
    auth: BasicAuth,
    scan_ref: str,
    action: Literal["pause", "cancel", "resume", "fetch", "delete"],
    **kwargs,
) -> Union[str, DataFrame]:
    """
    Perform an action on a VMDR scan.

    Args:
        auth (BasicAuth): Qualys BasicAuth object.
        scan_ref (str): The scan reference ID. Formatted like scan/1234567890123456.
        action (Literal["pause", "resume", "cancel", "fetch", "delete"]): The action to perform on the scan.

    :Keyword Args:
    NOTE: ALL KWARGS ARE FOR action="fetch" ONLY

        ips (str): The IPs to fetch. Defaults to None.
        mode (Literal["brief", "extended"]): The fetch mode. Defaults to "brief". If you specify nothing, or a non-existent mode, the SDK will default to "brief".
        client_id (int): The client ID. Defaults to None. Only available for consultant subscriptions.
        client_name (str): The client name. Defaults to None. Only available for consultant subscriptions.

    Returns:
        Union[str, pd.DataFrame]: Either a text response or a pandas DF.
    """

    # Check for valid action:
    if action not in ["pause", "resume", "cancel", "fetch", "delete"]:
        raise ValueError(
            f"Invalid action {action}. Must be one of 'pause', 'resume', 'cancel', or 'fetch'."
        )

    # Base payload for all actions:
    payload = {
        "scan_ref": scan_ref,
        "action": action,
        "echo_request": False,
    }

    # If action is "fetch", add the kwargs to the payload:
    if action == "fetch":
        match kwargs.get("mode"):
            case "brief":
                payload["output_format"] = "json"
            case "extended":
                payload["output_format"] = "json_extended"
            case _:
                payload["output_format"] = "json"

        payload.update(kwargs)

    # Make the request:
    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint=f"{action}_scan",  # Nifty way to dynamically get required params/payload for each action
        payload=payload,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    # Parse the response:

    # Special case for fetch.
    if action != "fetch":
        result = xml_parser(response.text)

        # Check for empty results:
        if not result:
            print("No scan paused.")
            return None

        data = result["SIMPLE_RETURN"]["RESPONSE"]

        if "CODE" in data.keys():
            raise QualysAPIError(data["TEXT"])

        return data["TEXT"]

    else:
        # Make sure that the response is JSON:
        if response.headers["Content-Type"] not in [
            "application/json",
            "application/json; charset=UTF-8",
            "text/html; charset=UTF-8",
        ]:
            raise QualysAPIError(
                xml_parser(response.text)["SIMPLE_RETURN"]["RESPONSE"]["TEXT"]
            )

        # Parse the JSON response:
        result = read_json(StringIO(response.text))

        if result.empty:
            print("No scan found.")
            return None

        return result


def pause_scan(auth: BasicAuth, scan_ref: str) -> str:
    """
    Pause a VMDR scan.

    Args:
        auth (BasicAuth): Qualys BasicAuth object.
        scan_ref (str): The scan reference ID. Formatted like scan/1234567890123456.

    Returns:
        str: The text response from Qualys
    """

    # Call manage_scan with the action set to "pause":
    return manage_scan(auth=auth, scan_ref=scan_ref, action="pause")


def resume_scan(auth: BasicAuth, scan_ref: str) -> str:
    """
    Resume a VMDR scan.

    Args:
        auth (BasicAuth): Qualys BasicAuth object.
        scan_ref (str): The scan reference ID. Formatted like scan/1234567890123456.

    Returns:
        str: The text response  from Qualys
    """

    # Call manage_scan with the action set to "resume":
    return manage_scan(auth=auth, scan_ref=scan_ref, action="resume")


def cancel_scan(auth: BasicAuth, scan_ref: str) -> str:
    """
    Cancel a VMDR scan.

    Args:
        auth (BasicAuth): Qualys BasicAuth object.
        scan_ref (str): The scan reference ID. Formatted like scan/1234567890123456.

    Returns:
        str: The text response from Qualys
    """

    # Call manage_scan with the action set to "cancel":
    return manage_scan(auth=auth, scan_ref=scan_ref, action="cancel")


def delete_scan(auth: BasicAuth, scan_ref: str) -> str:
    """
    Delete a VMDR scan.

    Args:
        auth (BasicAuth): Qualys BasicAuth object.
        scan_ref (str): The scan reference ID. Formatted like scan/1234567890123456.

    Returns:
        str: The text response from Qualys
    """

    # Call manage_scan with the action set to "delete":
    return manage_scan(auth=auth, scan_ref=scan_ref, action="delete")


def fetch_scan(auth: BasicAuth, scan_ref: str, **kwargs) -> DataFrame:
    """
    Fetch VMDR scan results.

    Args:
        auth (BasicAuth): Qualys BasicAuth object.
        scan_ref (str): The scan reference ID. Formatted like scan/1234567890123456.

    :kwargs:
        ips (str): The IPs to fetch. Defaults to None.
        mode (Literal["brief", "extended"]): The fetch mode. Defaults to "brief". If you specify nothing, or a non-existent mode, the SDK will default to "brief".
        client_id (int): The client ID. Defaults to None. Only available for consultant subscriptions.
        client_name (str): The client name. Defaults to None. Only available for consultant subscriptions.

    Returns:
        pd.DataFrame: The scan results in a pandas DataFrame.
    """

    # Call manage_scan with the action set to "fetch":
    return manage_scan(auth=auth, scan_ref=scan_ref, action="fetch", **kwargs)



"""
scanner_appliances.py - Contains functions to interact with Scanner Appliances in VMDR.
"""

from ..exceptions import *
from ..base import call_api, xml_parser
from .data_classes import ScannerAppliance
from .data_classes.lists import BaseList
from ..auth import BasicAuth


def get_scanner_list(auth: BasicAuth, **kwargs) -> BaseList[ScannerAppliance]:
    """
    Get a list of Scanner Appliances in VMDR, filtered by kwargs.

    Parameters:
        auth (BasicAuth): The BasicAuth context to use.
        **kwargs: Additional keyword arguments to filter the Scanner Appliances by.

    :Kwargs:
        action (str): The action to filter by. NOTE: the SDK force-sets this to "list". It is just included for completeness.
        echo_request (bool): Whether to echo the request. NOTE: the SDK force-sets this to False. It is just included for completeness.
        output_mode (Literal['brief', 'full']): The output mode. Defaults to 'brief'.
        scan_detail (bool): Whether to include details on currently running scans. Defaults to False. Details include ID, title, scan_ref, type, and date.
        show_tags (bool): Whether to include tags. Defaults to False.
        include_cloud_info (bool): Whether to include cloud information. Defaults to False. Requires output_mode='full'. If set, output includes cloud data for appliances in a cloud platform.
        busy (bool): Whether to include busy appliances. Defaults to True.
        scan_ref (str): The scan reference to filter by. Output will only show scanners that are running this scan.
        name (str): The name to filter by.
        ids (Union[str, int]): The IDs to filter by. Can be a single ID or a comma-separated string of IDs.
        include_license_info (bool): Whether to include license usage info. Output is under ['LICENSE_INFO']. NOTE: the SDK force-sets this to False. It is just included for completeness.
        type (Literal['physical', 'virtual', 'containerized', 'offline']): The type of scanner to filter by. Appears when output_mode='full'.
        platform_provider (Literal['ec2', 'ec2_compat', 'gce', 'azure', 'vCenter']): The cloud platform provider to filter by.

    Returns:
        BaseList[ScannerAppliance]: A list of Scanner Appliances.
    """

    result = BaseList()

    # set the action and echo_request:
    kwargs["action"] = "list"
    kwargs["echo_request"] = False
    kwargs["include_license_info"] = False

    # call the API:
    resp = call_api(
        auth=auth,
        module="vmdr",
        endpoint="get_scanner_list",
        params=kwargs,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    data = xml_parser(resp.text)["APPLIANCE_LIST_OUTPUT"]["RESPONSE"]

    if "APPLIANCE_LIST" not in data.keys():
        print("No data in response")
        return result

    # If there is just one dict, convert it to a list of one dict
    if isinstance(data["APPLIANCE_LIST"]["APPLIANCE"], dict):
        data = [data]

    for item in data["APPLIANCE_LIST"]["APPLIANCE"]:
        result.append(ScannerAppliance(**item))

    return result



"""
detection.py - contains the Detection dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *
from datetime import datetime
from warnings import catch_warnings, simplefilter

from bs4 import BeautifulSoup

from .qds_factor import QDSFactor
from .qds import QDS as qds
from ..data_classes.lists import BaseList


@dataclass(order=True)
class Detection:
    """
    Detection - represents a single QID detection on a host.
    """

    UNIQUE_VULN_ID: int = field(
        metadata={"description": "The unique ID of the detection."}, compare=False
    )
    QID: int = field(metadata={"description": "The QID of the detection."})
    TYPE: Literal["Confirmed", "Potential"] = field(
        metadata={"description": "The type of the detection."}
    )
    SEVERITY: int = field(metadata={"description": "The severity of the detection."})
    STATUS: Literal["New", "Active", "Fixed", "Re-Opened"] = field(
        metadata={"description": "The status of the detection."}, compare=False
    )
    SSL: Optional[bool] = field(
        metadata={"description": "The SSL status of the detection."},
        default=False,
        compare=False,
    )
    RESULTS: Optional[str] = field(
        metadata={"description": "The results of the detection."},
        default="",
        compare=False,
    )

    FIRST_FOUND_DATETIME: Union[str, datetime] = field(
        metadata={"description": "The date and time the detection was first found."},
        default=None,
        compare=False,
    )
    LAST_FOUND_DATETIME: Union[str, datetime] = field(
        metadata={"description": "The date and time the detection was last found."},
        default=None,
        compare=False,
    )

    QDS: Optional[qds] = field(
        metadata={"description": "The Qualys Detection Score (QDS) of the detection."},
        default=None,
        compare=False,
    )

    QDS_FACTORS: Optional[List[QDSFactor]] = field(
        metadata={
            "description": "The Qualys Detection Score (QDS) factors of the detection."
        },
        default=None,
        compare=False,
    )

    TIMES_FOUND: int = field(
        metadata={"description": "The number of times the detection was found."},
        default=0,
        compare=False,
    )
    LAST_TEST_DATETIME: Union[str, datetime] = field(
        metadata={"description": "The date and time the detection was last tested."},
        default=None,
        compare=False,
    )
    LAST_UPDATE_DATETIME: Union[str, datetime] = field(
        metadata={"description": "The date and time the detection was last updated."},
        default=None,
        compare=False,
    )
    IS_IGNORED: bool = field(
        metadata={"description": "The ignored status of the detection."},
        default=False,
        compare=False,
    )
    IS_DISABLED: bool = field(
        metadata={"description": "The disabled status of the detection."},
        default=False,
        compare=False,
    )
    LAST_PROCESSED_DATETIME: Union[str, datetime] = field(
        metadata={"description": "The date and time the detection was last processed."},
        default=None,
        compare=False,
    )
    LAST_FIXED_DATETIME: Optional[Union[str, datetime]] = field(
        metadata={"description": "The date and time the detection was last fixed."},
        default=None,
        compare=False,
    )
    PORT: Optional[int] = field(
        metadata={"description": "The port of the detection."},
        default=None,
        compare=False,
    )
    PROTOCOL: Optional[str] = field(
        metadata={"description": "The protocol of the detection."},
        default=None,
        compare=False,
    )
    FQDN: Optional[str] = field(
        metadata={"description": "The fully qualified domain name of the detection."},
        default=None,
        compare=False,
    )

    ID: int = field(
        metadata={"description": "The host ID of host the detection is on."},
        default=None,
    )

    def __post_init__(self):
        # convert the datetimes to datetime objects
        DATETIME_FIELDS = [
            "FIRST_FOUND_DATETIME",
            "LAST_FOUND_DATETIME",
            "LAST_TEST_DATETIME",
            "LAST_UPDATE_DATETIME",
            "LAST_PROCESSED_DATETIME",
            "LAST_FIXED_DATETIME",
        ]

        HTML_FIELDS = ["RESULTS"]

        BOOL_FIELDS = ["IS_IGNORED", "IS_DISABLED", "SSL"]

        INT_FIELDS = ["UNIQUE_VULN_ID", "QID", "SEVERITY", "TIMES_FOUND", "PORT", "ID"]

        for field in DATETIME_FIELDS:
            if (
                isinstance(getattr(self, field), str)
                and getattr(self, field) is not None
            ):
                setattr(self, field, datetime.fromisoformat(getattr(self, field)))

        # clean up fields that have html tags
        with catch_warnings():
            simplefilter("ignore")  # ignore the warning about the html.parser
            for field in HTML_FIELDS:
                setattr(
                    self,
                    field,
                    BeautifulSoup(getattr(self, field), "html.parser").get_text(),
                )

        # convert the BOOL_FIELDS to bool
        for field in BOOL_FIELDS:
            if (
                not isinstance(getattr(self, field), bool)
                and getattr(self, field) is not None
            ):
                setattr(self, field, bool(getattr(self, field)))

        # convert the INT_FIELDS to int
        for field in INT_FIELDS:
            if (
                not isinstance(getattr(self, field), int)
                and getattr(self, field) is not None
            ):
                setattr(self, field, int(getattr(self, field)))

        # convert the QDS to a QDS object
        if self.QDS:
            self.QDS = qds(SEVERITY=self.QDS["@severity"], SCORE=int(self.QDS["#text"]))

        # convert the QDS factors to QDSFactor objects
        if self.QDS_FACTORS:
            factors_bl = BaseList()
            data = self.QDS_FACTORS["QDS_FACTOR"]

            # Normalize QDS factors to a list for easier processing
            if isinstance(data, dict):
                data = [data]

            for factor in data:
                factors_bl.append(
                    QDSFactor(NAME=factor["@name"], VALUE=factor["#text"])
                )

            self.QDS_FACTORS = factors_bl

    def __str__(self):
        # return str(self.UNIQUE_VULN_ID)
        return str(self.QID)

    def __int__(self):
        return self.QID

    def copy(self):
        return Detection(
            UNIQUE_VULN_ID=self.UNIQUE_VULN_ID,
            QID=self.QID,
            TYPE=self.TYPE,
            SEVERITY=self.SEVERITY,
            SSL=self.SSL,
            RESULTS=self.RESULTS,
            STATUS=self.STATUS,
            FIRST_FOUND_DATETIME=self.FIRST_FOUND_DATETIME,
            LAST_FOUND_DATETIME=self.LAST_FOUND_DATETIME,
            TIMES_FOUND=self.TIMES_FOUND,
            LAST_TEST_DATETIME=self.LAST_TEST_DATETIME,
            LAST_UPDATE_DATETIME=self.LAST_UPDATE_DATETIME,
            LAST_FIXED_DATETIME=self.LAST_FIXED_DATETIME,
            IS_IGNORED=self.IS_IGNORED,
            IS_DISABLED=self.IS_DISABLED,
            LAST_PROCESSED_DATETIME=self.LAST_PROCESSED_DATETIME,
            QDS=self.QDS,
            QDS_FACTORS=self.QDS_FACTORS,
            PORT=self.PORT,
            PROTOCOL=self.PROTOCOL,
        )

    def valid_values(self):
        # return a list of attribute names that have non-None values
        return {k: v for k, v in self.items() if v is not None and v != "" and v != []}

    def to_dict(self):
        """
        to_dict - convert the Detection object to a dictionary.

        This function is used to convert the Detection object to a dictionary.
        """
        return asdict(self)

    def __dict__(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: Union[dict, list]):
        """
        from_dict - create a Software object from a dictionary.

        This function is used to create a Software object from a dictionary.
        """
        # make sure that the dictionary has the required keys and nothing else:
        required_keys = {"QID", "SEVERITY", "STATUS", "TYPE"}

        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        return cls(**data)



"""
SQL server integration for qualysdk
"""

from .base import db_connect
from .vmdr import (
    upload_vmdr_ags,
    upload_vmdr_kb,
    upload_vmdr_hosts,
    upload_vmdr_ips,
    upload_vmdr_hld,
    upload_vmdr_scanners,
    upload_vmdr_static_searchlists,
    upload_vmdr_users,
    upload_vmdr_scanlist,
)



"""
vendor_reference.py - contains the VendorReference dataclass for the Qualys VMDR module.
"""

from dataclasses import dataclass, field, asdict
from typing import *


@dataclass
class VendorReference:
    """
    VendorReference - represents a single vendor bulletin reference in a ReferenceList.

    This class is used to represent a single vendor bulletin reference in a ReferenceList,
    which is used to represent the vendor bulletin references for a vulnerability.
    """

    ID: str = field(metadata={"description": "The ID of the vendor reference."})
    URL: str = field(
        metadata={"description": "The URL of the vendor reference."},
        default="",
        compare=False,
    )

    def __post_init__(self):
        # check that url is a string:
        if not isinstance(self.URL, str):
            raise TypeError(
                f"VendorReference URL must be a string, not {type(self.URL)}"
            )

    def __str__(self) -> str:
        return self.ID

    def __contains__(self, item):
        # see if it was found in the name or vendor:
        return item in self.ID or item in self.URL

    def copy(self):
        return VendorReference(ID=self.ID, URL=self.URL)

    def is_id(self, id: int):
        return self.ID == id

    def is_url(self, url: str):
        return self.URL == url

    def to_dict(self):
        return asdict(self)

    def keys(self):
        return self.to_dict().keys()

    def values(self):
        return self.to_dict().values()

    def items(self):
        return self.to_dict().items()

    @classmethod
    def from_dict(cls, data: dict):
        """
        from_dict - create a Software object from a dictionary.

        This function is used to create a Software object from a dictionary.
        """
        # make sure that the dictionary has the required keys and nothing else:
        required_keys = {"ID"}
        if not required_keys.issubset(data.keys()):
            raise ValueError(
                f"Dictionary must contain the following keys: {required_keys}"
            )

        return cls(**data)



"""
vmscans.py - Contains the VMScan data class.
"""

from typing import Union, Literal
from datetime import datetime, timedelta
from dataclasses import dataclass, field, asdict

from .lists.base_list import BaseList
from .ip_converters import *


def parse_duration(duration: str) -> timedelta:
    """
    Parse a duration string into a timedelta object.

    Params:
        duration (str): Duration string in the format HH:MM:SS or N days HH:MM:SS.

    Returns:
        timedelta: Timedelta object representing the duration.
    """
    if "days" in duration:
        days, time = duration.split(" days ")
        days = int(days)
    else:
        days = 0
        time = duration

    try:
        hours, minutes, seconds = map(int, time.split(":"))
        return timedelta(days=days, hours=hours, minutes=minutes, seconds=seconds)
    except ValueError:
        return timedelta(days=days)


@dataclass(order=True)
class VMScan:
    """
    Data class to hold VM Scan data.
    """

    REF: str = field(
        metadata={
            "description": "Scan reference ID. Formatted like scan/1234567890123456, compliance/1234567890123456, or qscap/1234567890123456."
        },
        default=None,
    )
    TYPE: Literal["On-Demand", "API", "Scheduled"] = field(
        metadata={"description": "Type of scan. On-Demand, API, Scheduled."},
        default=None,
    )
    TITLE: str = field(metadata={"description": "Title of the scan."}, default=None)
    USER_LOGIN: str = field(
        metadata={"description": "Owner of the scan."}, default=None
    )
    LAUNCH_DATETIME: Union[str, datetime] = field(
        metadata={"description": "Datetime the scan was launched."}, default=None
    )  # post init to convert to datetime
    DURATION: Union[str, timedelta] = field(
        metadata={"description": "Duration of the scan."}, default=None
    )  # post init to convert to timedelta
    PROCESSING_PRIORITY: str = field(
        metadata={"description": "Processing priority of the scan."}, default=None
    )
    PROCESSED: bool = field(
        metadata={"description": "Whether the scan has been processed."}, default=None
    )
    STATUS: dict = field(
        metadata={"description": "Status of the scan."}, default=None
    )  # STATE is in this dict. get STATE in post init
    STATE: str = field(
        metadata={"description": "The state of the scan."}, init=False, default=None
    )  # populated in post init
    TARGET: Union[str, BaseList[str], BaseList[ip_address]] = field(
        metadata={"description": "Target of the scan."}, default=None
    )
    OPTION_PROFILE: dict = field(
        metadata={"description": "Option profile of the scan."}, default=None
    )
    ASSET_GROUP_TITLE_LIST: Union[str, BaseList[str]] = field(
        metadata={"description": "Asset group title list."}, default=None
    )

    def __post_init__(self):
        """
        Post init function to convert the LAUNCH_DATETIME and DURATION fields to datetime and timedelta objects.
        """
        if self.LAUNCH_DATETIME:
            self.LAUNCH_DATETIME = datetime.strptime(
                self.LAUNCH_DATETIME, "%Y-%m-%dT%H:%M:%SZ"
            )
        if self.DURATION:
            self.DURATION = (
                parse_duration(self.DURATION)
                if self.DURATION not in ["Pending", "Running"]
                else self.DURATION
            )

        if self.STATUS:
            self.STATE = self.STATUS["STATE"]

        # For the IPs/IP ranges in the TARGET field, determine if it is a single IP/range, or a comma separated string of IPs/ranges.
        if self.TARGET:
            final_list = BaseList()
            self.TARGET = self.TARGET.split(",")
            for t in self.TARGET:
                if "-" in t:
                    t = single_range(t)
                else:
                    t = single_ip(t)
                final_list.append(t)
            self.TARGET = final_list

        if self.PROCESSED:
            self.PROCESSED = bool(self.PROCESSED)

        # create the baselist for the asset group titles
        # first, check if [ASSET_GROUP_TITLE_LIST] is a dict. If it is, convert it to a list
        if self.ASSET_GROUP_TITLE_LIST:
            if isinstance(self.ASSET_GROUP_TITLE_LIST, dict):
                self.ASSET_GROUP_TITLE_LIST = [self.ASSET_GROUP_TITLE_LIST]
            # create the BaseList object
            final_list = BaseList()
            for ag in self.ASSET_GROUP_TITLE_LIST:
                final_list.append(ag["ASSET_GROUP_TITLE"])
            self.ASSET_GROUP_TITLE_LIST = final_list

    def to_dict(self) -> dict:
        """
        Convert the VMScan object to a dictionary.
        """
        return asdict(self)

    @classmethod
    def from_dict(cls, data: dict):
        """
        Create a VMScan object from a dictionary.
        """
        return cls(**data)

    def __str__(self):
        return f"VMScan: {self.TITLE} - {self.REF}"

    def __int__(self):
        return int(self.REF.split("/")[-1])

    def keys(self):
        return self.__dict__().keys()

    def values(self):
        return self.__dict__().values()

    def items(self):
        return self.__dict__().items()

    def valid_values(self):
        return {k: v for k, v in self.__dict__().items() if v}



"""
ip_converters.py - contains the IP and IPRange helpers for the VMDR module.

The helpers convert string inputs from raw API responses into ipaddress.IPvxAddress/IPvxNetwork objects.
"""

from ipaddress import (
    ip_address,
    ip_network,
    IPv4Address,
    IPv6Address,
    IPv4Network,
    IPv6Network,
    summarize_address_range,
)
from typing import Union


def single_ip(ip: str) -> ip_address:
    """
    Converts a single IP string into an ipaddress.IPv4Address or ipaddress.IPv6Address object.

    Params:
        ip (str): IP address string.

    Returns:
        Union[IPv4Address, IPv6Address]: IPv4Address or IPv6Address object.
    """
    return ip_address(ip)


def single_range(ip_range: str) -> ip_network:
    """
    Converts an IP range string into an ipaddress.IPv4Network or ipaddress.IPv6Network object.

    Params:
        ip_range (str): IP range string.

    Returns:
        Union[IPv4Network, IPv6Network]: IPv4Network or IPv6Network object.
    """
    start, end = ip_range.split("-")

    start, end = ip_address(start), ip_address(end)

    # A little confusing, but summarize_address_range returns a generator, so we use list comprehension and then grab the first element.
    return ip_network([i for i in summarize_address_range(start, end)][0])


# Bulk IP and IP range conversion functions:


def convert_ips(ips: list[str]) -> list[Union[IPv4Address, IPv6Address]]:
    """
    Converts a list of IP strings into a list of ipaddress.IPv4Address or ipaddress.IPv6Address objects.

    Params:
        ips (list[str]): List of IP address strings.

    Returns:
        list[Union[IPv4Address, IPv6Address]]: List of IPv4Address or IPv6Address objects.
    """
    return [single_ip(ip) for ip in ips]


def convert_ranges(ip_ranges: list[str]) -> list[Union[IPv4Network, IPv6Network]]:
    """
    Converts a list of IP range strings into a list of ipaddress.IPv4Network or ipaddress.IPv6Network objects.

    Params:
        ip_ranges (list[str]): List of IP range strings.

    Returns:
        list[Union[IPv4Network, IPv6Network]]: List of IPv4Network or IPv6Network objects.
    """
    return [single_range(ip_range) for ip_range in ip_ranges]



"""
users.py - contains functions to interact with VMDR user management APIs.
"""

from typing import List, Union, Literal

from ..auth import BasicAuth
from ..base import xml_parser, call_api
from .data_classes.lists import BaseList
from .data_classes.user import User
from ..exceptions import *


def get_user_list(auth: BasicAuth, **kwargs) -> BaseList[User]:
    """
    Get a list of users in the subscription.

    Args:
        auth (BasicAuth): The authentication object.

    :Kwargs:
        external_id__contains (str): Filter by external ID containing this string.
        external_id_assigned (bool): Filter by external ID assigned.

    Returns:
        BaseList[User]: BaseList of User objects.
    """

    response = call_api(
        auth=auth,
        endpoint="get_user_list",
        module="vmdr",
        params=kwargs,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    bl = BaseList()

    if response.status_code != 200:
        raise QualysAPIError(response.text)

    user_list = xml_parser(response.text)

    if "ERROR" in user_list["USER_LIST_OUTPUT"].keys():
        raise QualysAPIError(user_list["USER_LIST_OUTPUT"]["ERROR"]["#text"])

    if "USER" not in user_list["USER_LIST_OUTPUT"]["USER_LIST"].keys():
        print("No users found.")
        return bl

    # Check for single user
    if isinstance(user_list["USER_LIST_OUTPUT"]["USER_LIST"]["USER"], dict):
        user_list["USER_LIST_OUTPUT"]["USER_LIST"]["USER"] = [
            user_list["USER_LIST_OUTPUT"]["USER_LIST"]["USER"]
        ]

    for user in user_list["USER_LIST_OUTPUT"]["USER_LIST"]["USER"]:
        bl.append(User(**user))

    return bl


def add_user(
    auth: BasicAuth,
    user_role: Literal[
        "manager", "unit_manager", "scanner", "reader", "contact", "administrator"
    ],
    business_unit: Union[Literal["Unassigned"], str],
    first_name: str,
    last_name: str,
    title: str,
    phone: str,
    email: str,
    address1: str,
    city: str,
    country: str,
    state: str,
    **kwargs,
) -> str:
    """
    Adds a user to the subscription.

    Args:
        auth (BasicAuth): The authentication object.
        user_role (Literal["manager", "unit_manager", "scanner", "reader", "contact", "administrator"]): The role of the user.
        business_unit (Union[Literal["Unassigned"], str]): The business unit of the user.
        first_name (str): The first name of the user.
        last_name (str): The last name of the user.
        title (str): The title of the user.
        phone (str): The phone number of the user.

        email (str): The email of the user.
        address1 (str): The address of the user.
        city (str): The city of the user.
        country (str): The country of the user.
        state (str): The state of the user.

    :Kwargs:
        send_email (bool): Send an email to the user about their activation. True by default. If False, output will contain the username and password.
        asset_groups (str): Comma-separated string of asset group IDs.
        fax (str): The fax number of the user. Because 21st century...
        address2 (str): The second line of the address.
        zip_code (str): The zip code.
        external_id (str): The external ID of the user.

    Returns:
        str: The message from Qualys. If send_email is False, the output will contain the username and password.
    """

    params = {
        "action": "add",
        "user_role": user_role,
        "business_unit": business_unit,
        "first_name": first_name,
        "last_name": last_name,
        "title": title,
        "phone": phone,
        "email": email,
        "address1": address1,
        "city": city,
        "country": country,
        "state": state,
    }

    params.update(kwargs)

    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="add_user",
        params=params,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    if response.status_code != 200:
        raise QualysAPIError(response.text)

    result = xml_parser(response.text)

    if result["USER_OUTPUT"]["RETURN"]["@status"] == "FAILED":
        raise QualysAPIError(result["USER_OUTPUT"]["RETURN"]["MESSAGE"])

    if kwargs.get("send_email") == False:
        return f"User created. User:Pass is: {result['USER_OUTPUT']['USER']['USER_LOGIN']}:{result['USER_OUTPUT']['USER']['PASSWORD']}"

    else:
        return result["USER_OUTPUT"]["RETURN"]["MESSAGE"]


def edit_user(auth: BasicAuth, login: str, **kwargs) -> str:
    """
    Change details of a pre-existing user in Qualys.
    NOTE: To clear a field, specify it with an empty string in the kwargs.

    Args:
        auth (BasicAuth): The authentication object.
        login (str): The login of the user to edit.

    :Kwargs:
        asset_groups (str): Comma-separated string of asset group IDs.
        first_name (str): The first name of the user.
        last_name (str): The last name of the user.
        title (str): The title of the user.
        phone (str): The phone number of the user.
        fax (str): The fax number of the user. Because 21st century...
        email (str): The email of the user.
        address1 (str): The address of the user.
        address2 (str): The second line of the address.
        city (str): The city of the user.
        country (str): The country of the user.
        state (str): The state of the user.
        zip_code (str): The zip code.
        external_id (str): The external ID of the user.

    Returns:
        str: The message from Qualys.
    """

    params = {"action": "edit", "login": login}
    params.update(kwargs)

    response = call_api(
        auth=auth,
        module="vmdr",
        endpoint="edit_user",
        params=params,
        headers={"X-Requested-With": "qualysdk SDK"},
    )

    if response.status_code != 200:
        raise QualysAPIError(response.text)

    result = xml_parser(response.text)

    if result["USER_OUTPUT"]["RETURN"]["@status"] == "FAILED":
        raise QualysAPIError(result["USER_OUTPUT"]["RETURN"]["MESSAGE"])

    return result["USER_OUTPUT"]["RETURN"]["MESSAGE"]
